import socket
import time
import subprocess
import json
import logging
import re
from AudioSource import *
from threading import *
from informacast import InformaCastConfig as InformaCastConfig
from informacast import InformaCastClient as InformaCastClient
import baco

# ========================================================================================================
#    InformacastClient
#
#    1 - IPSpeaker CFG file loaded localy (by main-informacast process)
#    2 - Idle audio received by "start_idle_audio" command
#    3 - InformacastClient sets its "attention to Announcement" when receiving "start" command from Server
#    4 - InformacastClient sets its "attention to Idle Audio" when receiving "stop" command from Server
#    5 - Audio actions:
#            -> plays & monitor stream  (announce or idle audio) when is STATE is PLAYING
#            -> monitor stream  (announce or idle audio) when is STATE is MONITORING
#            -> no play/monitor when is STATE is STOPPED
# ========================================================================================================


class InformacastAudioSource(AudioSource):

    def __init__(self, sourceId, sourceName, sourceVolume, sourceConfig, audioDev):
        AudioSource.__init__(self, sourceId, sourceName, sourceVolume, sourceConfig, audioDev)

        # Information provided by Informacast Server (after Client Registration) to receive commands
        # self.command_port       = 12346
        self.local_sock_port = 9999
        self.local_sock_addr = "127.0.0.1"
        self.remote_sock_addr = None

        self.playing = False
        self.externalCallback = None
        self.command = ""  # initial condition

        self.maxICastVolume = 12 # range: 0 to 12

        self.announce_address = None
        self.announce_port = None
        self.announce_vol = 0
        self.announce_start = 0
        self.announce_alive = False
        # self.url_announce       = ""

        # Idle Audio info that should be retrieved from CFG file
        self.idle_address = None  # "239.1.1.1"
        self.idle_port = None  # 5004
        self.idle_vol = 0
        self.idle_alive = False
        # self.url_idle           = "" # "rtp://"+self.idle_address+":"+str(self.idle_port)

        self.audioDevice = audioDev

        # baco players
        self.playerAnnounce = None
        self.playerIdleAudio = None

        self.monitoringIdleAudioThread = None
        self.monitoringAnnounceThread = None
        self.isMonitoringIdle = False
        self.isMonitoringAnnounce = False

        self.cleared = False

        self.audio_mon = None

        self.logger = logging.getLogger('ipspeaker')

        self.running = False
        self.lock = Lock()

        # self.setVolume()
        self.sourceVol = sourceVolume
        self.prevVolume = self.sourceVol

        # report status
        self.counter = 0
        self.nbroadcastsPlayed = 0
        self.nbroadcastsRecv = 0
        self.registeredStatus = {}
        self.reg_interval = None
        self.fileConfigs = {}

        self.configure()

        self.start()

    def configure(self):
        # create UDP socket to receive commands from main-informacast client
        self.sock_command = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock_command.bind((self.local_sock_addr, self.local_sock_port))

        url = str(AudioSource.getParameter(self, "url"))
        """
        serverProtocol  = str(AudioSource.getParameter(self,"protocol"))
        serverIp        = str(AudioSource.getParameter(self,"server_address"))
        serverPort      = str(AudioSource.getParameter(self,"port"))
        """

        print(url)

        infocast_conf = None

        if url == "":
            infocast_conf = InformaCastConfig.InformaCastConfig(PlayerUDPPort=self.local_sock_port)
        else:
            idle_url = str(AudioSource.getParameter(self, "idleaudio"))
            idle_addr = None
            idle_port = None
            idle_format = None
            if idle_url != "":
                match = re.search(re.compile(r"^rtp://(\d+)\.(\d+)\.(\d+)\.(\d+)(:|@)(\d+)$"), idle_url)
                idle_addr = "{}.{}.{}.{}".format(match.group(1), match.group(2), match.group(3), match.group(4))
                idle_port = int(match.group(6))
                idle_format = str(AudioSource.getParameter(self, "idleaudio_format"))
            infocast_conf = InformaCastConfig.InformaCastConfig(PlayerUDPPort=self.local_sock_port, ServerURL=url,
                                                                IdleAddr=idle_addr, IdlePort=idle_port,
                                                                IdleFormat=idle_format,
                                                                IdleVol=int(int(self.sourceVol) / 10))

        self.informacastClient = InformaCastClient.InformaCastDevice(infocast_conf)

    def start(self):
        self.lock.acquire()
        try:
            if self.running == False:
                # initialize Thread for receiving Commands
                self.informacast_thread = Thread(target=self.rcv_command)
                self.informacast_thread.daemon = True
                self.stopRcvCommands = False
                self.informacast_thread.start()
                self.running = True
        finally:
            self.lock.release()

    def convICastVol(self, infvol):
        """
        INPUT: ICast volume range : 0 - 12
        OUPUT: IPAC volume range: 0 - 100
        """
        if infvol > self.maxICastVolume:
            self.maxICastVolume = infvol
        outVol = round(infvol*100/self.maxICastVolume)
        self.logger.info("ICast Vol ({}) => IPAC Vol ({})".format(infvol,outVol))
        return outVol

    def startIdleAudioPlayer(self, addr, port, vol):
        self.setVolume(vol)
        self.prevVolume = vol
        url = "rtp://{}:{}".format(addr, port)
        self.playerIdleAudio = baco.SimplePlayer(url, self.audioDevice, 44100)
        self.playerIdleAudio.setInputFifoDuration(0)
        self.playerIdleAudio.setOutputFifoDuration(0)
        self.playerIdleAudio.setCallback(self.checkIdleAudioStateCallback)
        self.playerIdleAudio.connectSource()

    def checkIdleAudioStateCallback(self, alive):
        self.idle_alive = alive
        if alive:
            self.logger.info("INFORMACAST: Idle Audio Source is ALIVE")
        else:
            self.logger.info("INFORMACAST: Idle Audio Source is DEAD")
        self.reportAliveStatus()

    def startAnnouncePlayer(self, addr, port, vol):
        self.setVolume(vol)
        self.prevVolume = vol
        url = "rtp://{}:{}".format(addr, port)
        self.playerAnnounce = baco.SimplePlayer(url, self.audioDevice, 44100)
        self.playerAnnounce.setInputFifoDuration(0)
        self.playerAnnounce.setOutputFifoDuration(0)
        self.playerAnnounce.setCallback(self.checkAnnounceStateCallback)
        self.logger.info("INFORMACAST: startAnnouncePlayer -> will connect Announce Source")
        self.playerAnnounce.connectSource()

    def checkAnnounceStateCallback(self, alive):
        self.announce_alive = alive
        if alive:
            self.logger.info("INFORMACAST: Announce Source is ALIVE")
        else:
            self.logger.info("INFORMACAST: Announce Source is DEAD")
        self.reportAliveStatus()

    #report Alive status to AudioSourceManager based on Announce and Idle Audio alive status
    def reportAliveStatus(self):
        if self.idle_alive or self.announce_alive:
            self.logger.info("INFORMACAST: Tell AudioSourceManager that is ALIVE")
            if self.externalCallback is not None:
                self.externalCallback(self, True)
        else:
            self.logger.info("INFORMACAST: Tell AudioSourceManager that is DEAD")
            if self.externalCallback is not None:
                self.externalCallback(self, False)

    def setCallback(self, callback):
        self.externalCallback = callback

    def startMonitoringIdle(self):
        if not self.isMonitoringIdle:
            self.isMonitoringIdle = True
            self.monitoringIdleAudioThread = Thread(target=self.reconnectIdleSourceLoop)
            self.monitoringIdleAudioThread.start()

    def reconnectIdleSourceLoop(self):
        while self.isMonitoringIdle:

            isConnected = self.playerIdleAudio.isConnected()
            if not isConnected:
                res = self.playerIdleAudio.connectSource()
                if res == 0:
                    self.logger.info("Idle Audio source id={}, connection succeed!".format(self.id))
                #else:
                #    self.logger.info("Idle Audio source id={}, connection failed.".format(self.id))

            time.sleep(1)
        self.isMonitoringIdle = False

    def startMonitoringAnnounce(self):
        if not self.isMonitoringAnnounce:
            self.isMonitoringAnnounce = True
            self.monitoringAnnounceThread = Thread(target=self.reconnectAnnounceSourceLoop)
            self.monitoringAnnounceThread.start()

    def reconnectAnnounceSourceLoop(self):
        while self.isMonitoringAnnounce:

            #self.logger.info("INFORMACAST: reconnectAnnounceSourceLoop")
            isConnected = self.playerAnnounce.isConnected()
            if not isConnected:
                res = self.playerAnnounce.connectSource()
                if res == 0:
                    self.logger.info("Announce source id={}, connection succeed!".format(self.id))
                else:
                    self.logger.info("Announce source id={}, connection failed.".format(self.id))

            time.sleep(1)
        self.logger.info("End monitoring Announce!")
        self.isMonitoringAnnounce = False


    def rcv_command(self):
        while self.stopRcvCommands == False:
            data, self.remote_sock_addr = self.sock_command.recvfrom(1024)
            print(self.remote_sock_addr)
            json_content = json.loads(data)
            print(json_content)
            if json_content["type"] == "command":
                command_json = json_content["content"]
                cmd = str(command_json["command_name"])

                if cmd == "start_idle_audio":  # save info about idle stream
                    self.idle_address = str(command_json["stream_ip"])
                    self.idle_port = int(command_json["port"])
                    self.idle_vol = self.convICastVol(int(command_json["volume"]))
                    self.logger.info("INFORMACAST: got IDLE audio configuration - {}:{}".format(self.idle_address, self.idle_port))
                    self.logger.info("INFORMACAST: Will Connect to Idle Source")
                    self.startIdleAudioPlayer(self.idle_address, self.idle_port, self.idle_vol)
                    self.startMonitoringIdle()
                    if self.playing:
                        if self.playerAnnounce is None:
                            self.playerIdleAudio.connectAudio()
                        elif not self.playerAnnounce.isConnected():
                            self.playerIdleAudio.connectAudio()
                elif cmd == "start":  # save info about announcement stream
                    self.announce_address = str(command_json["stream_ip"])
                    self.announce_port = int(command_json["port"])
                    self.announce_vol = self.convICastVol(int(command_json["volume"]))
                    self.logger.info("INFORMACAST: got announce - command:{}, address:{}, port: {} ".format(cmd, self.announce_address, self.announce_port))

                    # send Error message to ICast Server if player has received a second start without receiving the first stop
                    #if self.playerAnnounce is not None:
                    #    # TODO - Send Error message to ICast Server
                    #    self.logger.info("INFORMACAST: Cannot start a new announce without stop the previous one")
                    #else:

                    #if the player is already playing an announcement ... stop it a play the new one
                    if self.playerAnnounce is not None:
                        self.isMonitoringAnnounce = False
                        self.playerAnnounce = None

                    # stop playing Idle Audio if it exist
                    if self.playerIdleAudio is not None:
                        self.playerIdleAudio.disconnectAudio()

                    if self.playing:
                        # stop playing Idle Audio if it exist
                        #if self.playerIdleAudio is not None:
                        #    self.playerIdleAudio.disconnectAudio()
                        self.logger.info("INFORMACAST: Will Start Announce Player - start - Player already in PLAYING state")
                        self.startAnnouncePlayer(self.announce_address, self.announce_port, self.announce_vol)
                        self.startMonitoringAnnounce()
                        self.playerAnnounce.connectAudio()
                        self.nbroadcastsPlayed += 1
                        self.announce_start = time.time()
                    else: # State MONITORING
                        self.logger.info("INFORMACAST: Will Start Announce Player - start - Player in MONITORING or STOPPED state")
                        self.startAnnouncePlayer(self.announce_address, self.announce_port, self.announce_vol)
                        self.startMonitoringAnnounce()

                elif cmd == "stop":
                    # check if it corresponds to the same stream was started by ICast
                    self.logger.info("INFORMACAST: got announce - command:{}, address:{}, port: {} ".format(cmd, command_json["stream_ip"], command_json["port"]))
                    if command_json["stream_ip"] == self.announce_address and command_json["port"] == self.announce_port:
                        self.logger.info("INFORMACAST: stop - stream ip:{}, port:{}".format(self.announce_address,self.announce_port))
                        self.isMonitoringAnnounce = False
                        if self.playerAnnounce is not None:
                            if self.playerAnnounce.isConnected():
                                self.logger.info("INFORMACAST: stop - will disconnect baco")
                                self.playerAnnounce.disconnect()
                        #time.sleep(1)
                        self.playerAnnounce = None
                        self.logger.info("INFORMACAST: stop - disconnected playerAnnounce")
                        #need to report status here because baco lib get stuck 3 seconds before declaring that the stream is dead
                        self.announce_alive = False
                        self.reportAliveStatus()


                        # TODO - Must send packet statistics after receiving a Stop command
                        # pkts_statistics = self.stopMonitoring()
                        # stopResp = self.buildStopResponse(command_json["stream_ip"], command_json["port"], pkts_statistics)
                        # self.sendResponseUDP(stopResp)

                        if self.playing and self.playerIdleAudio is not None:
                            #if there is Idle Audio play it
                            self.playerIdleAudio.connectAudio()
                    else:
                        # TODO - Send Error message to ICast Server
                        self.logger.info("INFORMACAST: Cannot stop stream that is not active -> Will send error to ICast Server")

                elif cmd == "start_identification_tone":  # test command
                    self.logger.info("INFORMACAST: got start_identification_tone")
                    volume = self.convICastVol(int(command_json["volume"]))
                    duration = command_json["duration_in_seconds"]  # int
                    self.informacastClient.testCmd(self.audioDev, duration, volume)
                    # self.info_proc = subprocess.Popen(["speaker-test", "-D", self.audioDev, "-t", "sine", "--frequency", "800", "-buffer", "10000", "--nloops", "1"], shell=False)
                    # self.info_proc.wait()
                    # self.setVolume(self.prevVolume)
                    continue
                elif cmd == "clear":  # clear command
                    # stop playing everything
                    if self.playerIdleAudio is not None:
                        self.playerIdleAudio = None
                    if self.playerAnnounce is not None:
                        self.playerAnnounce = None
                    self.announce_address = None
                    self.announce_port = None
                    self.cleared = True
                    # Notify Audio Source Manager that informaCast has nothing to play
                    if self.externalCallback is not None:
                        self.externalCallback(self, False)
                    self.command = cmd  # update command
                    continue

                self.command = cmd  # update command
            elif json_content["type"] == "registrationStatus":
                # print("Received Registration Servers List")
                # print(json_content["content"])
                self.registeredStatus = json_content["content"]

            elif json_content["type"] == "registration-interval":
                self.reg_interval = json_content["content"]

            elif json_content["type"] == "configFileReport":
                self.fileConfigs = json_content["content"]

            elif json_content["type"] == "nbroadcastsRecv":
                self.nbroadcastsRecv = json_content["content"]

    def connect(self):
        pass

    def disconnect(self):
        self.stopAudio()

    def shutdown(self):
        with self.lock:
            self.logger.info(f"Shutdown source id={self.id}, name={self.name}")
            if self.running == True:
                if self.audio_mon is not None:
                    self.audio_mon.disconnect()
                if self.playing:
                    if self.playerIdleAudio is not None:
                        self.isMonitoringIdle = False
                        self.playerIdleAudio.disconnect()
                    if self.playerAnnounce is not None:
                        self.isMonitoringAnnounce = False
                        self.playerAnnounce.disconnect()
                self.playing = False
                self.stopRcvCommands = True
                self.sock_command.close()
                self.running = False
            self.informacastClient.disconnect()

    def playAudio(self):
        with self.lock:
            self.logger.info(f"Playing source id={self.id}, name={self.name}")
            try:
                if not self.playing:
                    if self.command == "start":
                        self.playerAnnounce.connectAudio()
                        self.nbroadcastsPlayed += 1
                        self.announce_start = time.time()

                    elif self.command == "stop" and not self.cleared:
                        self.playerIdleAudio.connectAudio()

                    elif self.command == "start_idle_audio" and not self.cleared:
                        if self.playerAnnounce is None:
                            self.playerIdleAudio.connectAudio()
                    else:
                        self.logger.warning("Unexpected order to play (last command received: {}".format(self.command))

                self.playing = True
            except Exception as e:
                self.logger.error(f"Play Audio Exception: {e}")

    def stopAudio(self):
        with self.lock:
            self.logger.info(f"Muting source id={self.id}, name={self.name}")
            try:
                if self.playerIdleAudio is not None:
                    self.playerIdleAudio.disconnectAudio()
                if self.playerAnnounce is not None:
                    self.playerAnnounce.disconnectAudio()
                self.playing = False
            except Exception as e:
                self.logger.error(f"Stop Audio Exception: {e}")

    def buildStopResponse(self, stream_ip, stream_port, jsonDict):
        print(jsonDict)
        print(type(jsonDict))
        stopResp = {'stream': stream_ip, 'port': stream_port}
        for key, value in jsonDict.items():
            stopResp[key] = value
        print(stopResp)
        return stopResp

    def sendResponseUDP(self, msg):
        print(msg)
        packet = json.dumps(msg).encode('utf-8')
        self.sock_command.sendto(packet, (self.remote_sock_addr))

    def reportStatus(self):
        sourceDict = {}
        sourceDict["id"] = self.id
        if self.command == "start":
            sourceDict["playing"] = "announcement"
        elif self.command == (
                self.command == "stop" or self.command == "start_idle_audio") and not self.cleared and self.idle_address is not None and self.idle_port is not None:
            sourceDict["playing"] = "background music"
        else:
            sourceDict["playing"] = "nothing"

        sourceDict["registered"] = self.registeredStatus
        sourceDict["registration_interval"] = self.reg_interval if self.reg_interval is not None else "unknown"
        sourceDict["configuration"] = self.fileConfigs  # if file is not downloaded, it goes empty

        idleDict = {}
        if self.idle_address is not None and self.idle_port is not None:
            idleDict["addr"] = self.idle_address
            idleDict["server_port"] = self.idle_port
        sourceDict["idleAudio"] = idleDict

        curBroadcast = {}
        if self.command == "start":
            curBroadcast["addr"] = self.announce_address
            curBroadcast["port"] = self.announce_port
            curBroadcast["startTime"] = self.announce_start * 1000  # ms
        sourceDict["curBroadcast"] = curBroadcast
        sourceDict["nBroadcastsRecv"] = self.nbroadcastsRecv
        sourceDict["nPlayedBroadcasts"] = self.nbroadcastsPlayed

        return sourceDict
