"""
    Local SchedulerAudioSource
    Scheduling of:
    - Fixed Time Announcements - High priority
    - Rotational Announcements - Medium priority
    - Playlist (music tracks) - Low priority
        - play in music tracks in order or shuffle them
"""

import logging
import time
import os
import sys
import json
import enum
import queue
import random
import alsaaudio
from datetime import datetime, timedelta
from threading import *

from AudioSource import *
import baco

logger = logging.getLogger('ipspeaker')

class SchedulerType(enum.IntEnum):
    FIXED = 0
    ROTATIONAL = 1
    PLAYLIST  = 2

class PlaylistTrigger(enum.IntEnum):
    """
    Used by scheduler to trigger playlist to start/stop playing
    """
    STOP = 0
    PLAY = 1

class Scheduler:
    def __init__(self):
        self.schedulerFile = "/mnt/data/config/scheduler.json"
        self.schedulerConfig = None
        self.schedulerTriggerList = []
        self.fixedTimeAnnounceList = []
        self.rotationalAnnounceList = []
        self.rotationalFileList = []
        self.tempRotationalFileList = []
        self.rotationalFileIndexToPlay = 0
        self.rotationalInterval = None
        self.rotationalDailyRestart = False
        self.listOfPlaylists = []
        self.tempListOfPlaylists = []
        self.triggerPlaylistState = []
        self.tempTriggerPlaylistState = []
        self.triggerCallback = None

        self.lock = Lock()

    def start(self):
        self.loadConfig()

        self.schedulerTriggerList = self.fixedTimeAnnounceList
        self.schedulerTriggerList.extend(self.rotationalAnnounceList)
        self.schedulerTriggerList = sorted(self.schedulerTriggerList, key=lambda x: (x['slot'], x['type']))
        self.rotationalFileList = self.tempRotationalFileList
        self.triggerPlaylistState = self.tempTriggerPlaylistState
        self.listOfPlaylists = sorted(self.tempListOfPlaylists, key=lambda x: x['end_slot'])
        print(f"playlist LIST:\n {self.listOfPlaylists}")
        print(f"triggerPlaylistState:\n {self.triggerPlaylistState}")

        self.stopWorkerThread = False
        self.workerThread = Thread(target=self._workedThreadFunction)
        self.workerThread.start()

    def loadConfig(self):
        """
        load schedule configuration
        """
        print("Load new scheduler config")
        self.schedulerConfig = self._loadSchedulerConfig()
        self.fixedTimeAnnounceList = self._createFixedTimeAnnouncementsList()
        self.rotationalAnnounceList, self.tempRotationalFileList = self._createRotationalAnnouncementsList()
        self.tempListOfPlaylists = self._createPlaylistConfigList()
        self.tempTriggerPlaylistState = self.setPlaylistScheduleTriggerState()

    def applyNewConfig(self):
        """
        apply new configuration
        """
        print("Apply new scheduler config")
        self.schedulerTriggerList = self.fixedTimeAnnounceList
        self.schedulerTriggerList.extend(self.rotationalAnnounceList)
        self.schedulerTriggerList = sorted(self.schedulerTriggerList, key=lambda x: (x['slot'], x['type']))
        self.rotationalFileList = self.tempRotationalFileList
        print(f"triggerPlaylistState\n{self.triggerPlaylistState}\n")
        print(f"TempTriggerPlaylistState\n{self.tempTriggerPlaylistState}\n")
        self.updateTempTriggerPlaylistState()
        print(f"TempTriggerPlaylistState\n{self.tempTriggerPlaylistState}\n")
        self.triggerPlaylistState = self.tempTriggerPlaylistState
        self.listOfPlaylists = sorted(self.tempListOfPlaylists, key=lambda x: x['end_slot'])

    def setPlaylistScheduleTriggerState(self):
        triggerState = []
        if len(self.tempListOfPlaylists):
            for playlist in self.tempListOfPlaylists:
                list = {'id':playlist['id'],'week_day':playlist['week_day'],'trigger_state':PlaylistTrigger.STOP.value}
                triggerState.append(list)
        return triggerState

    def getPlaylistTriggerState(self, id, week_day):
        """
        Get current trigger playlist state
        :param id: playlist id
        :param week_day : day of week (0 - 6)
        :return: trigger state
        """
        state = PlaylistTrigger.STOP.value
        for playlist in self.triggerPlaylistState:
            if playlist['id'] == id and playlist['week_day'] == week_day:
                state = playlist['trigger_state']
                break
        return state

    def setPlaylistTriggerState(self, id, week_day, state):
        """
        Set trigger playlist state
        :param id: playlist id
        :param week_day : day of week (0 - 6)
        :param state: playlist new state
        """
        for playlist in self.triggerPlaylistState:
            if playlist['id'] == id and playlist['week_day'] == week_day:
                playlist['trigger_state'] = state
                break

    def updateTempTriggerPlaylistState(self):
        if len(self.triggerPlaylistState) and len(self.tempTriggerPlaylistState):
            for currTrigger in self.triggerPlaylistState:
                for tempTrigger in self.tempTriggerPlaylistState:
                    if currTrigger['id'] == tempTrigger['id'] and currTrigger['week_day'] == tempTrigger['week_day']:
                        tempTrigger['trigger_state'] = currTrigger['trigger_state']
                        print(f"temp_trigger_state:{tempTrigger['trigger_state']}, curr_trigger_state:{currTrigger['trigger_state']}")

    def checkIfPlaylistInScheduler(self, playlistId):
        """
        Check if playslist Id is in the scheduler
        :param playlistId: playlist Id
        :return: True or False
        """
        if len(self.listOfPlaylists) > 0:
            for elem in self.listOfPlaylists:
                if playlistId == elem['id']:
                    print(f"Found playlistId:{playlistId} in scheduler list!")
                    return True
        return False

    def checkCurrentDayInPlaylistScheduler(self, playlistId):
        """
        Check if scheduled playlist should run on the current week day
        :param playlistId: playlist Id
        :return: True or False
        """
        curWeekDay = time.localtime().tm_wday
        if len(self.listOfPlaylists) > 0:
            for elem in self.listOfPlaylists:
                if playlistId == elem['id'] and curWeekDay == elem['week_day']:
                    print(f"PlaylistId:{playlistId} is schedule for week_day:{curWeekDay}!")
                    return True
        return False

    def sendTriggerCallback(self, callback):
        self.triggerCallback = callback

    def _loadSchedulerConfig(self):
        if os.path.isfile(self.schedulerFile):
            with open(self.schedulerFile) as jsonFile:
                return json.load(jsonFile)
        else:
            return ""

    def _calcWeekThresholdSlot(self, weekDayString, hour, min):
        """
        Calculate threshold slot to trigger play of file
        The week starts at Mondays ( 0 - Mon, 1 - Tue, ... 6 - Sun)
        :param weekDayString: week day string (ex: Mon, Tue, ... Sat, Sun)
        :param hour: hour
        :param min: minutes
        :return: threshold slot
        """
        weekDay = time.strptime(weekDayString, "%a").tm_wday
        return weekDay*24*3600 + hour*3600 + min*60

    def _calcCurrentWeekSlot(self):
        """
        Calculate current week Slot
        :return: current slot and weekday
        """
        now = time.localtime()
        return now.tm_wday*24*3600 + now.tm_hour*3600 + now.tm_min*60, now.tm_wday

    def _createFixedTimeAnnouncementsList(self):
        """
        Create a list of Fixed Time Announcements
        Set the threshold slot for each announcement
        :return list of Fixed Time Announcements
        """
        try:
            print(self.schedulerConfig)
            fixedTimeAnnounce = []
            if "fixed_time" in self.schedulerConfig:
                print(self.schedulerConfig["fixed_time"])
                if len(self.schedulerConfig["fixed_time"]):
                    for fixedAnnounce in self.schedulerConfig["fixed_time"]:
                        hour, minutes = fixedAnnounce["time"].split(":")
                        for weekDay in fixedAnnounce["weekdays"]:
                            fixedAnnounceDict = {}
                            fixedAnnounceDict["slot"] = self._calcWeekThresholdSlot(weekDay,int(hour),int(minutes))
                            fixedAnnounceDict["type"] = SchedulerType.FIXED.value
                            fixedAnnounceDict["file_id"] = fixedAnnounce["fileId"]
                            fixedAnnounceDict["time"] = fixedAnnounce["time"]
                            fixedAnnounceDict["week_day"] = time.strptime(weekDay, "%a").tm_wday
                            fixedTimeAnnounce.append(fixedAnnounceDict)
            #print(f"Fixed-time List: {fixedTimeAnnounce}")
        except Exception as e:
            logger.error(f"Exception creating fixed time announcements list: {e}")
        finally:
            return fixedTimeAnnounce

    def _createRotationalAnnouncementsList(self):
        """
        Create a list of Rotational Announcements
        Set the threshold slot for each announcement
        :return list of rotational announcements
        """
        try:
            rotationalAnnounce = []
            rotationalFiles = []
            if "rotational" in self.schedulerConfig:
                rotational = self.schedulerConfig["rotational"]
                self.rotationalDailyRestart = rotational["restart_daily"]
                intervalMeasure = rotational["interval_measure"]
                intervalTime = rotational["interval_time"]
                if intervalMeasure == "minutes":
                    self.rotationalInterval = intervalTime * 60
                elif intervalMeasure == "hours":
                    self.rotationalInterval = intervalTime * 3600
                print(f"rotationalInterval (sec): {self.rotationalInterval}")
                if "filesList" in rotational and len(rotational["filesList"]) and "working_time" in rotational:
                    for audioFile in rotational["filesList"]:
                        if audioFile["active"]:
                            rotationalFiles.append(audioFile)
                    print(f"rotational files list: {rotationalFiles}")
                    for weekDay in rotational["working_time"]:
                        print(rotational["working_time"][weekDay])
                        hour, minutes = rotational["working_time"][weekDay]["start_time"].split(":")
                        startTimestamp = int(hour)*3600 + int(minutes)*60
                        hour, minutes = rotational["working_time"][weekDay]["end_time"].split(":")
                        endTimestamp = int(hour)*3600 + int(minutes)*60
                        timeCounter = startTimestamp
                        print(f"{startTimestamp}, {endTimestamp}, {timeCounter}")
                        while timeCounter + self.rotationalInterval < endTimestamp:
                            hour, minutes, seconds = str(timedelta(seconds=(timeCounter))).split(":")
                            rotationalDayDict = {}
                            rotationalDayDict["slot"] = self._calcWeekThresholdSlot(weekDay, int(hour), int(minutes))
                            rotationalDayDict["type"] = SchedulerType.ROTATIONAL.value
                            rotationalDayDict["time"] = f"{hour}:{minutes}"
                            rotationalDayDict["week_day"] = time.strptime(weekDay, "%a").tm_wday
                            rotationalAnnounce.append(rotationalDayDict)
                            timeCounter += self.rotationalInterval
            #print(f"Rotational List: {rotationalAnnounce}")
            print(f"numRotationTriggers: {len(rotationalAnnounce)}")
            print("")
        except Exception as e:
            logger.error(f"Exception rotational time announcements list: {e}")
        finally:
            return rotationalAnnounce, rotationalFiles

    def _createPlaylistConfigList(self):
        """
        Create a list of Playlists with relevant information
        Set the threshold slot for playlist start time and stop time
        :return list of playlists
        """
        try:
            listOfPlaylist = []
            if "playlists" in self.schedulerConfig:
                if len(self.schedulerConfig["playlists"]):
                    for playlist in self.schedulerConfig["playlists"]:
                        playlistList = []
                        for weekDay in playlist["weekdays"]:
                            playlistWeekDayDict = {}
                            hour, minutes = playlist["start_time"].split(":")
                            playlistWeekDayDict["start_slot"] = self._calcWeekThresholdSlot(weekDay, int(hour), int(minutes))
                            hour, minutes = playlist["end_time"].split(":")
                            playlistWeekDayDict["end_slot"] = self._calcWeekThresholdSlot(weekDay, int(hour), int(minutes))
                            playlistWeekDayDict["type"] = SchedulerType.PLAYLIST.value
                            playlistWeekDayDict["id"] = playlist["id"]
                            playlistWeekDayDict["start_time"] = playlist["start_time"]
                            playlistWeekDayDict["end_time"] = playlist["end_time"]
                            playlistWeekDayDict["week_day"] = time.strptime(weekDay, "%a").tm_wday
                            playlistList.append(playlistWeekDayDict)
                        listOfPlaylist.extend(playlistList)
        except Exception as e:
            logger.error(f"Exception creating Playlist Config list: {e}")
        finally:
            return listOfPlaylist


    def _workedThreadFunction(self):
        print("Start Worker Thread")
        prevCurrentSlot = sys.maxsize
        prevWeekDay = None

        monotonicTimePrevRef = time.monotonic()
        localTimePrev = (datetime.now() - datetime(1970, 1, 1)) / timedelta(seconds=1)

        while not self.stopWorkerThread:

            with self.lock:
                # time jump check
                timeJumped = False
                time.tzset()
                monotonicTimeCurrRef = time.monotonic()
                localTimeCurr = (datetime.now() - datetime(1970, 1, 1)) / timedelta(seconds=1)
                refStep = int((monotonicTimeCurrRef - monotonicTimePrevRef)/60)
                localStep = int((localTimeCurr - localTimePrev)/60)
                #print(f"localStep:{localStep}, refStep:{refStep}")
                if localStep - refStep != 0:
                    timeJumped = True
                    print("TIME JUMPED!!")

                monotonicTimePrevRef = monotonicTimeCurrRef
                localTimePrev = localTimeCurr

                currentSlot, weekDay = self._calcCurrentWeekSlot()

                if prevWeekDay is not None:
                    if prevWeekDay != weekDay and self.rotationalDailyRestart:
                        print("Reset rotational file list")
                        self.rotationalFileIndexToPlay = 0
                prevWeekDay = weekDay

                # scheduler to trigger Announcements (Fixed/Rotational)
                for trigger in self.schedulerTriggerList:
                    if trigger['week_day'] == weekDay:
                        if currentSlot >= prevCurrentSlot:
                            if trigger["slot"] == currentSlot and prevCurrentSlot < trigger["slot"] and not timeJumped:
                                print(f"A - will play type:{trigger['type']}, time:{trigger['time']}")
                                if self.triggerCallback is not None:
                                    self.triggerCallback(trigger)
                        else: # start new week
                            if currentSlot == trigger["slot"] and not timeJumped:
                                print(f"B - will play type:{trigger['type']}, time:{trigger['time']}")
                                if self.triggerCallback is not None:
                                    self.triggerCallback(trigger)

                # schedule playlist
                # start trigger condition: start_slot <= currentSlot < end_slot and still not triggered to play
                # stop trigger condition: (currentSlot < start_slot or currentSlot >= end_slot) and triggered to play
                if self.triggerCallback is not None:
                    for playlist in self.listOfPlaylists:
                        state = self.getPlaylistTriggerState(playlist['id'],weekDay)
                        if playlist["week_day"] == weekDay:
                            print(f"currentSlot:{currentSlot}, start_slot:{playlist['start_slot']}, end_slot:{playlist['end_slot']}, state:{state}")
                            if currentSlot >= prevCurrentSlot:
                                if currentSlot >= playlist["start_slot"] and currentSlot < playlist["end_slot"] and state == PlaylistTrigger.STOP.value and not timeJumped:
                                    print(f"A - will start to play playlist id:{playlist['id']}, weekDay:{weekDay}")
                                    print(f"currentSlot:{currentSlot}, start_slot:{playlist['start_slot']}, stop_slot:{playlist['end_slot']}")
                                    self.setPlaylistTriggerState(playlist['id'],weekDay,PlaylistTrigger.PLAY.value)
                                    print(self.triggerPlaylistState)
                                    self.triggerCallback(playlist)
                                if ( currentSlot < playlist["start_slot"] or currentSlot >= playlist["end_slot"] ) and state == PlaylistTrigger.PLAY.value and not timeJumped:
                                    print(f"A - will stop playing playlist id:{playlist['id']}, weekDay:{weekDay}")
                                    print(f"currentSlot:{currentSlot}, start_slot:{playlist['start_slot']}, stop_slot:{playlist['end_slot']}")
                                    self.setPlaylistTriggerState(playlist['id'],weekDay,PlaylistTrigger.STOP.value)
                                    print(self.triggerPlaylistState)
                                    self.triggerCallback(playlist, False)

                            else: # start new week
                                if currentSlot >= playlist["start_slot"] and currentSlot < playlist["end_slot"] and state == PlaylistTrigger.STOP.value and not timeJumped:
                                    print(f"B -New week: will start to play playlist id :{playlist['id']}, weekDay:{weekDay}")
                                    self.setPlaylistTriggerState(playlist['id'],weekDay,PlaylistTrigger.PLAY.value)
                                    print(self.triggerPlaylistState)
                                    self.triggerCallback(playlist)
                                if (currentSlot < playlist["start_slot"] or currentSlot >= playlist["end_slot"])  and state == PlaylistTrigger.PLAY.value and not timeJumped:
                                    print(f"B - New week: will stop playing playlist id:{playlist['id']}, weekDay:{weekDay}")
                                    self.setPlaylistTriggerState(playlist['id'],weekDay,PlaylistTrigger.STOP.value)
                                    print(self.triggerPlaylistState)
                                    self.triggerCallback(playlist, False)

                prevCurrentSlot = currentSlot
            time.sleep(1)
        print("Exit Worker Thread")

    def getRotationalFileAttributes(self):
        if len(self.rotationalFileList)>0:
            rotationFileAttributes = self.rotationalFileList[self.rotationalFileIndexToPlay]
            print(f"Rotational File:{rotationFileAttributes}, index:{self.rotationalFileIndexToPlay}")
            self.rotationalFileIndexToPlay += 1
            if self.rotationalFileIndexToPlay == len(self.rotationalFileList):
                self.rotationalFileIndexToPlay = 0
            return rotationFileAttributes
        else:
            print("empty rotational file list")
            return ""

    def shutdown(self):
        print("Scheduler shutdown...")
        self.stopWorkerThread = True
        self.workerThread.join()


class PlaylistState(enum.IntEnum):
    """
    Current state reporting if there is some playlist playing or none playing
    Not related to a specific playlist!
    Used by scheduler audio source
    """
    STOPPED = 0
    PLAYING = 1


class SchedulerAudioSource(AudioSource):
    """
    Scheduler Audio Source
    - Receives triggers from Scheduler and plays then
    - Announcements (Fixed/Rotational)
        - triggers are stored in a queue (fixed should be played immediately)
        - 1 player to play Announcements
    - Playlists
        - will be handled separately from other events
        - 1 player to play Playlists
    """
    def __init__(self, sourceId, name, volume, parameters, audioDev):
        self.audioDevice = audioDev[0]
        self.audioDevice1 = audioDev[1]
        AudioSource.__init__(self, sourceId, name, volume, parameters, audioDev[0])
        volAnnounce = str(AudioSource.getParameter(self, "vol_announcement"))
        volPlaylist = str(AudioSource.getParameter(self, "vol_playlist"))
        conf = self.audioDevice.split(':')
        alsaaudio.Mixer(conf[1]).setvolume(int(volAnnounce))
        conf = self.audioDevice1.split(':')
        alsaaudio.Mixer(conf[1]).setvolume(int(volPlaylist))

        # local audio files/ ids
        self.audioIdList = "/mnt/data/config/audioFiles.json"
        self.audioFolder = "/mnt/data/AudioFiles"
        self.playlistIdList = "/mnt/data/config/playlists.json"

        self.lock = Lock()
        self.externalCallback = None

        # trigger queue for events ( except Playlist events )
        self.triggerQueue = None
        self.scheduler = None

        self.announcePlayer = None
        self.announceIsAlive = False

        self.announceAlivePlaylistMuted = False

        self.playlistPlayer = None
        self.playlistIsAlive = False
        self.playlistFilesList = []
        self.playlistName = ""
        self.playlistID   = None
        self.playlistState = PlaylistState.STOPPED
        self.playlistIsMuted = False
        self.playlistFileIndex = 0

        self.sourcePlaying = False
        self.currentAnnounceFilePlaying = ""
        self.currentPlaylistFileID = None
        self.currentPlaylistFilePlaying = ""

        self.stopRunning = False
        self.mainThread = None
        self.isSchedulerDisconnected = True

    def startScheduler(self):
        with self.lock:
            logger.info("Start a new scheduler")
            self.triggerQueue = queue.Queue()
            self.scheduler = Scheduler()
            self.scheduler.triggerCallback = self._processTrigger
            self.scheduler.start()
            self.stopRunning = False
            self.mainThread = Thread(target=self._runMainLoop)
            self.mainThread.start()

    def setSchedulerConfig(self):
        """
        Set new scheduler config
        """
        with self.lock:
            logger.info("Set new scheduler config")
            if self.scheduler is not None:
                self.scheduler.loadConfig()
                self.scheduler.applyNewConfig()
                if self.playlistID is not None:
                    if self.scheduler.checkIfPlaylistInScheduler(self.playlistID):
                        if self.scheduler.checkCurrentDayInPlaylistScheduler(self.playlistID):
                            # Current Playlist files update and check
                            self._updateCurrentPlaylist()
                            print(f"number of playlist Files:{len(self.playlistFilesList)}\nfiles:{self.playlistFilesList}")
                            if  self.playlistState == PlaylistState.PLAYING and self.playlistPlayer is not None:
                                if len(self.playlistFilesList) == 0:
                                    logger.info("Current playlist has no audio files -> disconnect playlist player!")
                                    self.playlistPlayer.disconnect()
                                    self.playlistState = PlaylistState.STOPPED
                                else:
                                    if self.currentPlaylistFileID not in self.playlistFilesList:
                                        print(f"{self.currentPlaylistFileID} is not longer available")
                                        logger.info("Current playing playlist file is no longer available... play next playlist audio file")
                                        self._playNextPlaylistFile()
                        else:
                            logger.info("Current playlist is not allow to play today -> disconnect playlist player")
                            if self.playlistPlayer is not None:
                                self.playlistPlayer.disconnect()
                            self.playlistState = PlaylistState.STOPPED
                    else:
                        logger.info(f"Current playlist {self.playlistID} - {self.playlistName} is not scheduled -> disconnect playlist player")
                        if self.playlistPlayer is not None:
                            self.playlistPlayer.disconnect()
                        self.playlistState = PlaylistState.STOPPED


    def _processTrigger(self, trigger, start=True):
        """
        Callback function that will process trigger
        fixed trigger -> play immediately
        rotational trigger -> add trigger to queue
        playlist trigger -> start/stop playlist
        :param trigger:
        """
        with self.lock:
            if trigger['type'] == SchedulerType.FIXED.value:
                self._playAnnounceTrigger(trigger)
            elif trigger['type'] == SchedulerType.ROTATIONAL.value:
                logger.info(f"Add Rotational trigger to queue")
                self.triggerQueue.put(trigger)
            elif trigger['type'] == SchedulerType.PLAYLIST.value:
                logger.info(f"Playlist trigger")
                if start:
                    self._startPlaylist(trigger)
                else:
                    self._stopPlaylist()
            else:
                logger.error("Invalid trigger!!")


    def _getAnnounceTrigger(self):
        """
        Get announce trigger from queue
        """
        if not self.triggerQueue.empty():
            return self.triggerQueue.get()
        else:
            return None

    def _getFileId(self, filename):
        """
        Retrieve file Id of filename
        :param filename: audio file name
        :return: file Id
        """
        try:
            res = ""
            with open(self.audioIdList) as f:
                data = json.load(f)
                for index in data:
                    if index["filename"] == filename:
                        return index["file_id"]
            return res
        except Exception as e:
            print(f"getFileId Exception {e}")
            return res

    def _getFilename(self, fileId):
        """
        Retrieve filename from file id
        :param fileId: audio file id
        :return: filename
        """
        try:
            res = ""
            with open(self.audioIdList) as f:
                data = json.load(f)
                for index in data:
                    if index["file_id"] == fileId:
                        return index["filename"]
            return res
        except Exception as e:
            print(f"getFileId Exception {e}")
            return res

    def _playAnnounceTrigger(self, trigger):
        """
        play any announce Trigger: Fixed or Rotational
        before playing announce it should mute any playing playlist
        """
        if self.playlistState == PlaylistState.PLAYING and self.playlistPlayer is not None:
            self.playlistPlayer.disconnectAudio()
            self.playlistIsMuted = True
            logger.info("Disconnect Playlist audio before playing the announce")

        if trigger['type'] == SchedulerType.FIXED.value:
            filename = self._getFilename(trigger['file_id'])
            if filename != "":
                logger.info(f"Fixed Announce trigger: {filename}")
                self.announcePlayer = baco.SimplePlayer(f"{self.audioFolder}/{trigger['file_id']}", self.audioDevice, 44100)
                self.announcePlayer.setCallback(self.checkAnnounceStateCallback)
                self.setEqualizer(self.announcePlayer)
                self.announcePlayer.connectSource()
                self.currentAnnounceFilePlaying = filename
                if self.sourcePlaying:
                    self.announcePlayer.connectAudio()
                    logger.info("Announce: connect audio")
            else:
                logger.error(f"Did not find filename for fileId:{trigger['filename']}!")
        elif trigger['type'] == SchedulerType.ROTATIONAL.value:
            if self.scheduler is not None:
                fileAttrib = self.scheduler.getRotationalFileAttributes()
                if fileAttrib != "" and fileAttrib['active']:
                    filename = self._getFilename(fileAttrib['fileId'])
                    if filename != "":
                        logger.info(f"Rotational Announce trigger: {filename}, active: {fileAttrib['active']}")
                        self.announcePlayer = baco.SimplePlayer(f"{self.audioFolder}/{fileAttrib['fileId']}", self.audioDevice, 44100)
                        self.announcePlayer.setCallback(self.checkAnnounceStateCallback)
                        self.setEqualizer(self.announcePlayer)
                        self.announcePlayer.connectSource()
                        self.currentAnnounceFilePlaying = filename
                        if self.sourcePlaying:
                            self.announcePlayer.connectAudio()
                            logger.info("Announce: connect audio")
                    else:
                        logger.error(f"filename {fileAttrib['filename']} has no file ID!")
        else:
            self.logger.error(f"Wrong announce trigger type:{trigger['type']}")

    def _startPlaylist(self, trigger):
        logger.info("Start Playlist Player")
        try:
            self.playlistID = trigger["id"]
            self.playlistFilesList, self.playlistName = self._getPlaylistFiles(trigger["id"])
            logger.info(f"playlistFilesList:{self.playlistFilesList}, name:{self.playlistName}")
            if len(self.playlistFilesList):
                self.playlistFileIndex = 0
                self._playlistPlayer()
        except Exception as e:
            logger.error(f"Exception start Playlist {e}")
        finally:
            self.playlistState = PlaylistState.PLAYING

    def _stopPlaylist(self):
        logger.info("Stop Playlist Player")
        try:
            if self.playlistPlayer is not None:
                self.playlistPlayer.disconnect()
        except Exception as e:
            logger.error(f"Exception stop Playlist {e}")
        finally:
            self.playlistState = PlaylistState.STOPPED

    def _getPlaylistFiles(self, playlistId):
        """
        Get playlist files list. Shuffle the list if requested
        :param playlistId:
        :return: playlist Files List and name of playlist
        """
        try:
            res = [] , ""
            with open(self.playlistIdList) as f:
                data = json.load(f)
                logger.info(f"_getPlaylistFiles data:{data}")
                if playlistId in data:
                    fileList = data[playlistId]["filesList"]
                    logger.info(f"fileList:{fileList}")
                    shuffle = data[playlistId]["shuffle"]
                    name = data[playlistId]["name"]
                    if len(fileList) >0 and shuffle:
                        random.shuffle(fileList)
                    res = fileList, name
                else:
                    logger.error(f"Playlist Id {playlistId} not found!")
            return res
        except Exception as e:
            print(f"getFileId Exception {e}")
            return res

    def _updateCurrentPlaylist(self):
        # update playlist files in case of active playlist
        if self.playlistID is not None:
            self.playlistFilesList, self.playlistName = self._getPlaylistFiles(self.playlistID)
            logger.info(f"playlistFilesList:{self.playlistFilesList}, name:{self.playlistName}")

    def _playlistPlayer(self):
        logger.info("_playlistPlayer start ...")
        self.currentPlaylistFileID = self.playlistFilesList[self.playlistFileIndex]
        logger.info("Before creating new playlist player")
        self.playlistPlayer = baco.SimplePlayer(f"{self.audioFolder}/{self.currentPlaylistFileID}", self.audioDevice1, 44100)
        logger.info("Created new playlist player")
        self.playlistPlayer.setCallback(self.checkPlaylistStateCallback)
        self.setEqualizer(self.playlistPlayer)
        self.playlistPlayer.connectSource()
        self.currentPlaylistFilePlaying = self._getFilename(self.playlistFilesList[self.playlistFileIndex])
        logger.info(f"_playlistPlayer: announceIsAlive:{self.announceIsAlive}, sourcePlaying:{self.sourcePlaying}")
        if not self.announceIsAlive and self.sourcePlaying:
            self.playlistPlayer.connectAudio()
            logger.info("Playlist: connect audio")
            #self.playlistIsMuted = False

    def _playNextPlaylistFile(self):
        if len(self.playlistFilesList):
            self.playlistFileIndex += 1
            if self.playlistFileIndex >= len(self.playlistFilesList):
                self.playlistFileIndex = 0
            logger.info(f"playlist: play next file: {self.playlistFilesList[self.playlistFileIndex]}, index:{self.playlistFileIndex}")
            self._playlistPlayer()
        else:
            logger.info("Empty playlist: nothing to play")

    def _runMainLoop(self):
        mutedCounter = 0
        while not self.stopRunning:
            # less priority triggers should be processed here!
            with self.lock:
                announceAlive = self.announceIsAlive
                playlistAlive = self.playlistIsAlive
                if not announceAlive:
                    trigger = self._getAnnounceTrigger()
                    if trigger is not None:
                        self._playAnnounceTrigger(trigger)
                    else:
                        # if playlist was muted to play an announcement it is time to unmute it
                        # this only be checked when a rotational announcement did not start - avoiding playing both at same time
                        if self.announceAlivePlaylistMuted and self.playlistState == PlaylistState.PLAYING and self.sourcePlaying:
                            if self.playlistPlayer is not None:
                                self.playlistPlayer.connectAudio()
                                logger.info("Announce not alive -> Playlist: connect audio")
                            self.playlistIsMuted = False
                            self.announceAlivePlaylistMuted = False

                        # if playlist is still muted and the start of announce never became alive ... we wait 3 seconds to start playing playlist again
                        # this will be a very unusual situation...
                        if self.playlistIsMuted :
                            mutedCounter += 1
                            if mutedCounter == 3 and self.playlistState == PlaylistState.PLAYING and self.sourcePlaying:
                                if self.playlistPlayer is not None:
                                    self.playlistPlayer.connectAudio()
                                    logger.warning("Announce never became alive (for 3 sec) -> Playlist: connect audio")
                            self.playlistIsMuted = False

                else: #reset counter
                    mutedCounter = 0

                # check if it is time to change playlist song and upload new scheduler config
                if not playlistAlive and self.playlistState == PlaylistState.PLAYING:
                    self._playNextPlaylistFile()

            time.sleep(1)

    def connect(self):
        if self.isSchedulerDisconnected:
            self.isSchedulerDisconnected = False
            logger.info(f"Connecting source id={self.id}, name={self.name}")
            self.startScheduler()

    def disconnect(self):
        with self.lock:
            logger.info(f"Disconnecting source id={self.id}, name={self.name}")
            if self.scheduler is not None:
                self.scheduler.shutdown()
            if self.announcePlayer is not None:
                self.announcePlayer.disconnect()
            if self.playlistPlayer is not None:
                self.playlistPlayer.disconnect()
                self.playlistState = PlaylistState.STOPPED
            self.stopRunning = True
            self.sourcePlaying = False
            self.isSchedulerDisconnected = True

    def shutdown(self):
        self.logger.info("Shutdown Scheduler audio source")
        self.disconnect()

    def playAudio(self):
        logger.info(f"Playing source id={self.id}, name={self.name}")
        logger.info(f"announceIsAlive: {self.announceIsAlive}, playlistIsAlive: {self.playlistIsAlive}")
        if self.announceIsAlive:
            if self.announcePlayer is not None:
                self.announcePlayer.connectAudio()
                logger.info("Announce: connect audio")
        elif self.playlistIsAlive and self.playlistPlayer is not None:
            self.playlistPlayer.connectAudio()
            self.playlistIsMuted = False
            logger.info("Playlist: connect audio")

        self.sourcePlaying = True

    def stopAudio(self):
        logger.info(f"Muting source id={self.id}, name={self.name}")
        if self.announcePlayer is not None:
            self.announcePlayer.disconnectAudio()
            logger.info("Announce: disconnect audio")
        if  self.playlistPlayer is not None:
            self.playlistPlayer.disconnectAudio()
            logger.info("Playlist: disconnect audio")
            #self.playlistIsMuted = True

        self.sourcePlaying = False

    def reportStatus(self):
        sourceDict = {}
        sourceDict["id"] = self.id
        if self.announceIsAlive:
            sourceDict["scheduler"] = "Announcement"
            sourceDict["playing"] = self.currentAnnounceFilePlaying
        elif self.playlistIsAlive:
            sourceDict["scheduler"] = "Playlist"
            sourceDict["playlist_name"] = self.playlistName
            sourceDict["playing"] = self.currentPlaylistFilePlaying
        return sourceDict

    def setCallback(self, callback):
        self.externalCallback = callback

    def checkAnnounceStateCallback(self, alive):
        if alive:
            logger.info("Media Announcement is alive")
            if not self.announceIsAlive:
                logger.warning("announce become alive!")
                self.announceAlivePlaylistMuted = True
        else:
            logger.info("Media Announcement is dead")
        self.announceIsAlive = alive
        self.reportAliveStatus()

    def checkPlaylistStateCallback(self, alive):
        self.playlistIsAlive = alive
        if alive:
            logger.info("Media from Playlist is alive")
        else:
            logger.info("Media from Playlist is dead")
        self.reportAliveStatus()

    def reportAliveStatus(self):
        """
        Report Alive status to AudioSourceManager based on Announce and Playlist alive status
        """
        if self.announceIsAlive or self.playlistIsAlive:
            logger.info("Tell AudioSourceManager that is ALIVE")
            if self.externalCallback is not None:
                self.externalCallback(self, True)
        else:
            if self.playlistState == PlaylistState.STOPPED:
                logger.info("Tell AudioSourceManager that is DEAD")
                if self.externalCallback is not None:
                    self.externalCallback(self, False)

    def setEqualizer(self, player):
        AudioSource.setEqualizer(self, player)

    def setEqualizerParameter(self, config):
        if self.announceIsAlive and self.announcePlayer is not None:
            AudioSource.setEqualizerParameter(self, self.announcePlayer, config)
        if self.playlistState == PlaylistState.PLAYING and self.playlistPlayer:
            AudioSource.setEqualizerParameter(self, self.playlistPlayer, config)
        return True

    def enableEqualizer(self, enable):
        if self.announceIsAlive and self.announcePlayer is not None:
            AudioSource.enableEqualizer(self, self.announcePlayer, enable)
        if self.playlistState == PlaylistState.PLAYING and self.playlistPlayer:
            AudioSource.enableEqualizer(self, self.playlistPlayer, enable)

if __name__ == "__main__":
    scheduler = Scheduler()


