"""
REST API for IPAC
"""
import logging
import json
from functools import wraps
import alsaaudio
import datetime
import os
import socket
from enum import Enum
import subprocess
import shlex
import sys
import flask
from flask import Blueprint, request, Response
from apiUtils import *
from barix.web.uci import getValueOfUci, setUciConfigs
from barix.web.system_info import *
from barix.web.audio_files import uploadAudioFile
from barix import HTTPRequestsAPI
sys.path.append("../new_webui")
from ipformer import *
from audio_files import *
from restart_services import createListOfServicesToRestart

bp = Blueprint('restAPI', __name__)

log = logging.getLogger('flask-backend')

class Result(str, Enum):
    OK = 'OK'
    INV_PARAM = 'INVALID_PARAMETERS'
    INV_API_KEY = 'INVALID_API_KEY'
    NOT_CONFIG = 'NOT_CONFIGURED'
    FAIL_CURR_SW_INST = 'FAIL_CURR_SW_INST'
    FILE_NOT_FOUND = 'FILE_NOT_FOUND'
    ALREADY_PLAYING = 'ALREADY_PLAYING'
    NOT_ALLOWED = 'NOT_ALLOWED'


API_KEY_FILE_PATH = "/mnt/data/config/api_key.json"

def apiKeyRequired(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        apiKeyAuth = getValueOfUci("application", "main_config", "api_auth")
        if apiKeyAuth == "none":
            return f(*args, **kwargs)
        else:
            try:
                apiKey = None
                if os.path.isfile(API_KEY_FILE_PATH):
                    with open(API_KEY_FILE_PATH) as jsonFile:
                        apiDict = json.load(jsonFile)
                        apiKey = apiDict['api-key']
                else:
                    log.error("API Key file not found")
                    return Response(status=500)

                if apiKey is None:
                    log.error(f"API key not found in {API_KEY_FILE_PATH}")
                    return Response(status=500)

                reqApiKey = request.headers.get('X-API-Key')
                if apiKey == reqApiKey:
                    return f (*args, **kwargs)
                else:
                    return Response(json.dumps({"result":Result.INV_API_KEY}), status=401, headers={"Content-Type": "application/json"})
            except Exception as e:
                log.error(e)
                return Response(status=500)
    return wrapper



@bp.route('/api/v1/version', methods=['GET'])
@apiKeyRequired
def getImageVersion():
    res = {}
    try:
        res['version'] = getAppVersion()
    except Exception as e:
        log.error(e)
        return Response(status=500)

    return Response(json.dumps(res), status=200,headers={"content-type": "application/json"})


@bp.route('/api/v1/volume', methods=['GET','PUT'])
@apiKeyRequired
def masterVolume():
    res = {}
    try:
        if flask.request.method == 'GET':
            res['volume'] = int(getValueOfUci("application","audio","volume"))
            return Response(json.dumps(res), status=200,headers={"content-type": "application/json"})
        else: # PUT
            volCmd = json.loads(request.data)
            if "volume" in volCmd:
                vol = volCmd['volume']
                if getValueOfUci("application", "audio", "muted") == 'false':
                    mixer = alsaaudio.Mixer('BARIX_MAIN')
                    mixer.setvolume(vol)
                setUciConfigs({'application.audio.volume': f"{vol}"},commit=True,restartServices=False)
                return Response(json.dumps({"result":Result.OK}),status=200,headers={"content-type": "application/json"})
            else:
                return Response(json.dumps({"result":Result.INV_PARAM}),status=400,headers={"content-type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/mute', methods=['PUT'])
@apiKeyRequired
def mute():
    try:
        muteCmd = json.loads(request.data)
        if "mute" in muteCmd and type(muteCmd['mute']) == bool:
            mute = "false"
            if muteCmd['mute']:
                mute = "true"

            muteUciCfg = getValueOfUci("application", "audio", "muted")
            currVolume = int(getValueOfUci("application","audio","volume"))
            if muteUciCfg != mute:
                mixer = alsaaudio.Mixer('BARIX_MAIN')
                if muteCmd['mute']:
                    mixer.setvolume(0)
                else:
                    mixer.setvolume(currVolume)
                setUciConfigs({'application.audio.muted': mute},commit=True,restartServices=False)
            return Response(json.dumps({"result":Result.OK}),status=200,headers={"content-type": "application/json"})
        else:
            return Response(json.dumps({"result":Result.INV_PARAM}),status=400,headers={"content-type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/sonicip', methods=['GET','PUT'])
@apiKeyRequired
def sonicIP():
    try:
        if flask.request.method == 'GET':
            enabled = False
            if getValueOfUci("sonic_ip", "general", "enabled") == 'true':
                enabled = True
            res = {
                "enabled" : enabled,
                "volume" : int(getValueOfUci("sonic_ip","general","volume"))
            }
            return Response(json.dumps(res), status=200, headers={"content-type": "application/json"})
        else: # PUT
            sonicIpCmd = json.loads(request.data)
            if "enabled" in sonicIpCmd and type(sonicIpCmd['enabled']) == bool:
                setUciConfigs({'sonic_ip.general.enabled': sonicIpCmd['enabled']}, commit=True, restartServices=False)
            else:
                return Response(json.dumps({"result":Result.INV_PARAM}),status=400, headers={"content-type": "application/json"})

            if "volume" in sonicIpCmd:
                setUciConfigs({'sonic_ip.general.volume': f"{int(sonicIpCmd['volume'])}"},commit=True,restartServices=False)
            else:
                return Response(json.dumps({"result":Result.INV_PARAM}),status=400, headers={"content-type": "application/json"})

            return Response(json.dumps({"result":Result.OK}),status=200, headers={"content-type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/reset', methods=['PUT'])
@apiKeyRequired
def resetDefaults():
    try:
        # check if there is no updater process running
        if os.path.isfile("/tmp/fw_update"):
            return Response(json.dumps({"result":Result.FAIL_CURR_SW_INST}),status=403,headers={"content-type": "application/json"})

        subprocess.Popen(shlex.split("/usr/bin/barix-restore-defaults.sh soft > /dev/null 2>&1"))
        return Response(json.dumps({"result":Result.OK}),status=200,headers={"content-type": "application/json"})
    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/reboot', methods=['PUT'])
@apiKeyRequired
def rebootDevice():
    try:
        # check if there is no updater process running
        if os.path.isfile("/tmp/fw_update"):
            return Response(json.dumps({"result": Result.FAIL_CURR_SW_INST}), status=403, headers={"content-type": "application/json"})

        subprocess.Popen(shlex.split("/sbin/reboot"))
        return Response(json.dumps({"result":Result.OK}),status=200,headers={"content-type": "application/json"})
    except Exception as e:
        log.error(e)
        return Response(status=500)

def getDnsSettings():
    dnsServers = getDnsServers()
    if str(dnsServers) != "Auto":
        dnsType = "Manual"
        if len(dnsServers) == 2:
            primaryDNS = dnsServers[0]
            secundaryDNS = dnsServers[1]
        else:
            primaryDNS = dnsServers[0]
            secundaryDNS = ""
        dns = {"type": dnsType, "primaryDNS" : primaryDNS, "secundaryDNS" : secundaryDNS}
    else:
        dnsType = "Auto"
        dns = {"type" : dnsType}

    return dns

def getUptime(currDatetime):
    bootTime = datetime.datetime.fromtimestamp(getUpTime())
    timeDiff = currDatetime - bootTime
    m, s = divmod(timeDiff.seconds, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)
    return f"{timeDiff.days} days, {h} hours, {m} minutes, {s} seconds"


@bp.route('/api/v1/status', methods=['GET'])
@apiKeyRequired
def status():
    try:
        deviceInfo = {
            "hardwareType" : f"{getDeviceType()} ({getDeviceTypeID()})",
            "ipamType" : f"{getIPAMType()} (ID {getIPAMTypeID()})",
            "mac" : getMACaddr(),
            "linuxKernelVersion" : getKernelVer(),
            "bootloaderVersion": getBootloaderVer()
        }

        dhcp = False
        if getValueOfUci("network", "eth0", "proto") == 'dhcp':
            dhcp = True

        network = {
            "connection" : "wired",
            "status" : getWiredConnStatus(),
            "dhcp": dhcp,
            "ipaddress" : getIPaddr("eth0"),
            "netmask": getNetmask(),
            "defaultGateway": getDefaultGW(),
            "dns" : getDnsSettings()
        }

        currDatetime = datetime.datetime.fromtimestamp(getSystemTime())

        system = {
            "time": currDatetime.strftime('%a %b %e %H:%M:%S UTC %Y'),
            "uptime": getUptime(currDatetime)
        }

        res = {
            "deviceInfo" : deviceInfo,
            "network" : network,
            "system" : system,
            "licenses": getLicenses()
        }

        return Response(json.dumps(res), status=200, headers={"content-type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/network', methods=['GET'])
@apiKeyRequired
def network():
    try:
        avahiAnnounce = False
        if getValueOfUci("avahi","general","enabled") == 'true':
            avahiAnnounce = True

        sonicIP = False
        if getValueOfUci("sonic_ip","general","enabled") == 'true':
            sonicIP = True

        dhcp = False
        if getValueOfUci("network","eth0","proto") == 'dhcp':
            dhcp = True

        res = {
            "avahiAnnounce": avahiAnnounce,
            "sonicIP": sonicIP,
            "sonicIpVolume": int(getValueOfUci("sonic_ip", "general", "volume")),
            "webProtocol": getValueOfUci("httpd","webserver","protocol"),
            "dhcp": dhcp,
            "dhcpHostName": getValueOfUci("network","eth0","dhcpname"),
            "ipaddress": getIPaddr("eth0"),
            "netmask": getNetmask(),
            "gateway": getDefaultGW(),
            "dns": getDnsSettings()
        }

        return Response(json.dumps(res), status=200, headers={"content-type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


 ################################################################################################
 ################################################################################################
 # IPAC API specific commands

@bp.route('/api/v1/sources', methods=['GET','PUT'])
@apiKeyRequired
def audioSources():
    try:
        if flask.request.method == 'GET':
            sources = getIPFormerSources()
            res = { "sources": sources}
            return Response(json.dumps(res), status=200, headers={"content-type": "application/json"})
        elif flask.request.method == 'PUT':
            if 'multipart/form-data' in request.content_type:
                if 'file' not in request.files:
                    return  Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
                file = request.files['file']
                if file.filename == '':
                    log.error("Missing file")
                    return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
                file.save("/tmp/sources.json")
                with open("/tmp/sources.json") as f:
                    try:
                        sourcesCfg = json.load(f)
                    except ValueError:
                        return Response(json.dumps({"result": Result.INV_PARAM, "description": "Invalid json file"}), status=400, headers={"Content-Type": "application/json"})

            elif 'application/json' in request.content_type:
                sourcesCfg = json.loads(request.data)
            else:
                log.error("Invalid content-type for set audio sources PUT request method")
                return Response(json.dumps({"result": Result.INV_PARAM, "description": "Invalid request content-type"}), status=400, headers={"Content-Type": "application/json"})

        valid, description = validateSourcesConfiguration(sourcesCfg)
        if not valid:
            return Response(json.dumps({"result": Result.INV_PARAM, "description": description}), status=400, headers={"Content-Type": "application/json"})

        setIpFormerSources(sourcesCfg['sources'])

        return Response(json.dumps({"result": Result.OK}), status=200, headers={"Content-Type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/source/enable', methods=['PUT'])
@apiKeyRequired
def sourceEnable():
    try:
        sourceCmd = json.loads(request.data)
        if "id" in sourceCmd and "enable" in sourceCmd and type(sourceCmd["id"]) == int and type(sourceCmd["enable"]) == bool:
            if ( sourceCmd["id"] >= 1 ) and (sourceCmd["id"] <= len(getIPFormerSources()) ):
                #check current sourceid enable state
                sourceEnabled = getValueOfUci("flexa_ip_former", f"source{sourceCmd['id']}", "enabled")
                if sourceEnabled == "":
                    return Response(json.dumps({"result":Result.NOT_CONFIG}), status=400, headers={"Content-Type": "application/json"})

                enable = 'false'
                if sourceCmd["enable"] :
                    enable = 'true'

                setUciConfigs({f"flexa_ip_former.source{sourceCmd['id']}.enabled": enable}, commit=True, restartServices=True, selectServicesToRestartCallback=createListOfServicesToRestart)
                return Response(json.dumps({"result": Result.OK}), status=200,headers={"Content-Type": "application/json"})
            else:
                return Response(json.dumps({"result":Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
        else:
            return Response(json.dumps({"result":Result.INV_PARAM}),status=400, headers={"Content-Type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/source/volume', methods=['PUT'])
@apiKeyRequired
def sourceVolume():
    try:
        sourceCmd = json.loads(request.data)
        if "id" in sourceCmd and type(sourceCmd["id"]) == int:
            if ( sourceCmd["id"] >= 1 ) and (sourceCmd["id"] <= len(getIPFormerSources()) ):
                # check current sourceid enable state
                sourceEnabled = getValueOfUci("flexa_ip_former", f"source{sourceCmd['id']}", "enabled")
                if sourceEnabled == "":
                    return Response(json.dumps({"result": Result.NOT_CONFIG}), status=400, headers={"Content-Type": "application/json"})

                sourceType = getValueOfUci("flexa_ip_former", f"source{sourceCmd['id']}", "type")
                if sourceType == "scheduler":
                    schedulerCfg = json.loads(getValueOfUci("flexa_ip_former", f"source{sourceCmd['id']}", "config"))
                    if "vol_announcement" in sourceCmd and "vol_playlist" in sourceCmd and type(sourceCmd["vol_announcement"]) == int and type(sourceCmd["vol_announcement"]) == int:
                        volCtrl = alsaaudio.Mixer("chan_11")
                        volCtrl.setvolume(sourceCmd["vol_announcement"])
                        schedulerCfg["vol_announcement"] = sourceCmd["vol_announcement"]
                        volCtrl = alsaaudio.Mixer("chan_12")
                        volCtrl.setvolume(sourceCmd["vol_playlist"])
                        schedulerCfg["vol_playlist"] = sourceCmd["vol_playlist"]
                        setUciConfigs({f"flexa_ip_former.source{sourceCmd['id']}.config": json.dumps(schedulerCfg)}, commit=True, restartServices=False, selectServicesToRestartCallback=createListOfServicesToRestart)
                        return Response(json.dumps({"result": Result.OK}), status=200, headers={"Content-Type": "application/json"})
                    else:
                        return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})

                else: #common source with 1 volume
                    sourceVolume = getValueOfUci("flexa_ip_former", f"source{sourceCmd['id']}", "volume")
                    if sourceVolume == "":
                        return Response(json.dumps({"result":Result.NOT_CONFIG}), status=400, headers={"Content-Type": "application/json"})

                    if ( sourceCmd["volume"] >= 0 ) and ( sourceCmd["volume"] <= 100 ) :
                        volCtrl = alsaaudio.Mixer(f"chan_{sourceCmd['id']}")
                        volCtrl.setvolume(sourceCmd["volume"])
                        setUciConfigs({f"flexa_ip_former.source{sourceCmd['id']}.volume": sourceCmd['volume']}, commit=True, restartServices=False, selectServicesToRestartCallback=createListOfServicesToRestart)
                        return Response(json.dumps({"result":Result.OK}),status=200, headers={"Content-Type": "application/json"})
                    else:
                        return Response(json.dumps({"result":Result.INV_PARAM}),status=400, headers={"Content-Type": "application/json"})

            else:
                return Response(json.dumps({"result":Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
        else:
            return Response(json.dumps({"result":Result.INV_PARAM}),status=400, headers={"Content-Type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/sources/status', methods=['GET'])
@apiKeyRequired
def audioSourcesStatus():
    try:
        sourceStatus = getSourcesStatus()
        sourceStatus["status"].pop(0)
        return Response(json.dumps(sourceStatus), status=200, headers={"content-type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/audio-files', methods=['GET'])
@apiKeyRequired
def audioFiles():
    try:
        fileIdList = os.listdir("/mnt/data/AudioFiles")
        fileList = []
        for fileId in fileIdList:
            filename = getFilename(fileId)
            if filename != "":
                fileList.append(filename)
        if os.path.isdir("/run/media/sda1/triggered_files"):
            usbFileList = os.listdir("/run/media/sda1/triggered_files")
            fileList.extend(usbFileList)
        audioFiles = {"audio-files": fileList}
        return Response(json.dumps(audioFiles), status=200, headers={"content-type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)

@bp.route('/api/v1/file/upload', methods=['POST'])
@apiKeyRequired
def uploadFile():
    try:
        if 'file' not in request.files:
            return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
        file =  request.files['file']
        print(file)
        if file.filename == '':
            log.error("Missing file")
            return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
        audioFilesFormat = ["mp3", "aac", "flac", "ac3", "dts", "wav", "ogg", "m4a", "zip"]
        freeSpace = getStorageSize()["available"]
        uploadAudioFile(file,'/mnt/data/AudioFiles/',audioFilesFormat, freeSpace, AUDIO_FILES_JSON_FILE)
        return Response(json.dumps({"result": Result.OK}), status=200, headers={"Content-Type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)

@bp.route('/api/v1/file/delete', methods=['DELETE'])
@apiKeyRequired
def deleteFile():
    try:
        cmd = json.loads(request.data)
        if "filename" in cmd and type(cmd["filename"]) == str:
            audioFilesJson = "/mnt/data/config/audioFiles.json"
            with open(audioFilesJson) as jsonFile:
                data = json.load(jsonFile)
                for audioFile in data:
                    if audioFile["filename"] == cmd["filename"]:
                        deleteAudioFiles([audioFile['file_id']])
                        return Response(json.dumps({"result": Result.OK}), status=200, headers={"Content-Type": "application/json"})
            return Response(json.dumps({"result": Result.FILE_NOT_FOUND}), status=400, headers={"Content-Type": "application/json"})
        else:
            return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
    except Exception as e:
        log.error(e)
        return Response(status=500)

@bp.route('/api/v1/file/play', methods=['PUT'])
@apiKeyRequired
def filePlay():
    try:
        cmd = json.loads(request.data)
        if "id" in cmd and "filename" in cmd and type(cmd["id"]) == int and type(cmd["filename"]) == str:
            if cmd["id"] >= 1 and cmd["id"] <= len(getIPFormerSources()):
                #check if source id is a local_file audio source
                if getValueOfUci("flexa_ip_former", f"source{cmd['id']}", "type") != "local_file":
                    return Response(json.dumps({"result": Result.NOT_CONFIG}), status=400, headers={"Content-Type": "application/json"})
                sourceCfg = getValueOfUci("flexa_ip_former", f"source{cmd['id']}", "config")
                if "port" not in sourceCfg:
                    return Response(json.dumps({"result": Result.NOT_CONFIG}), status=400, headers={"Content-Type": "application/json"})

                # check for optional 'volume' parameter
                if "volume" in cmd:
                    vol = cmd['volume']
                    if type(cmd["volume"]) == int and vol >= 0 and vol <= 100:
                        log.info(f"set source{cmd['id']} volume to {vol}")
                        volCtrl = alsaaudio.Mixer(f"chan_{cmd['id']}")
                        volCtrl.setvolume(vol)
                    else:
                        log.error(f"Invalid volume value: {vol}")
                        return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})

                config = json.loads(sourceCfg)
                udpPort = int(config['port'])
                filename = cmd['filename']

                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                server_address = ("127.0.0.1", 60001)
                s.bind(server_address)
                s.settimeout(1.0)

                udpCmd = f"FILEPLAY={filename}\n"
                s.sendto(udpCmd.encode('utf-8'), ("127.0.0.1",udpPort))

                data, address = s.recvfrom(1024)
                resp = data.decode('utf-8').splitlines()

                if resp[0] == Result.OK:
                    return Response(json.dumps({"result": Result.OK}),status=200,headers={"Content-Type": "application/json"})
                elif resp[0] == Result.ALREADY_PLAYING:
                    return Response(json.dumps({"result": Result.ALREADY_PLAYING}), status=400, headers={"Content-Type": "application/json"})
                elif resp[0] == Result.FILE_NOT_FOUND:
                    return Response(json.dumps({"result": Result.FILE_NOT_FOUND}), status=400, headers={"Content-Type": "application/json"})
                elif resp[0] == "ERROR":
                    return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
            else:
                return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
        else:
            return Response(json.dumps({"result":Result.INV_PARAM}),status=400, headers={"Content-Type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/file/stop', methods=['PUT'])
@apiKeyRequired
def fileStop():
    try:
        cmd = json.loads(request.data)
        if "id" in cmd and type(cmd["id"]) == int:
            if cmd["id"] >= 1 and cmd["id"] <= len(getIPFormerSources()):
                #check if source id is a local_file audio source
                if getValueOfUci("flexa_ip_former", f"source{cmd['id']}", "type") != "local_file":
                    return Response(json.dumps({"result": Result.NOT_CONFIG}), status=400, headers={"Content-Type": "application/json"})
                sourceCfg = getValueOfUci("flexa_ip_former", f"source{cmd['id']}", "config")
                if "port" not in sourceCfg:
                    return Response(json.dumps({"result": Result.NOT_CONFIG}), status=400, headers={"Content-Type": "application/json"})

                config = json.loads(sourceCfg)
                udpPort = int(config['port'])

                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                server_address = ("127.0.0.1", 60002)
                s.bind(server_address)
                s.settimeout(1.0)

                udpCmd = "FILESTOP\n"
                s.sendto(udpCmd.encode('utf-8'), ("127.0.0.1",udpPort))

                data, address = s.recvfrom(1024)
                resp = data.decode('utf-8').splitlines()

                if resp[0] == "OK":
                    return Response(json.dumps({"result": Result.OK}),status=200,headers={"Content-Type": "application/json"})
                else:
                    return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
            else:
                return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
        else:
            return Response(json.dumps({"result":Result.INV_PARAM}),status=400, headers={"Content-Type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/file/status', methods=['GET'])
@apiKeyRequired
def fileStatus():
    try:
        cmd = json.loads(request.data)
        if "id" in cmd and type(cmd["id"]) == int:
            if cmd["id"] >= 1 and cmd["id"] <= len(getIPFormerSources()):
                if getValueOfUci("flexa_ip_former", f"source{cmd['id']}", "type") != "local_file":
                    return Response(json.dumps({"result": Result.NOT_CONFIG}), status=400, headers={"Content-Type": "application/json"})

                statusList = getSourcesStatus()
                resp = {}
                for source in statusList["status"]:
                    if source["id"] == cmd["id"] and source["type"] == "local_file":
                        if source["state"] == "playing":
                            resp = {"status": "playing", "name": source["playing"]}
                        else:
                            resp = {"status": "stopped"}
                        break

                return Response(json.dumps(resp), status=200, headers={"Content-Type": "application/json"})
            else:
                return Response(json.dumps({"result": Result.NOT_CONFIG}), status=400, headers={"Content-Type": "application/json"})
        else:
            return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)



@bp.route('/api/v1/inputs', methods=['GET'])
@apiKeyRequired
def inputsState():
    try:
        r = HTTPRequestsAPI.get('http://127.0.0.1:50500/api/inputs/get')
        info = r.read()
        enc = r.info().get_content_charset('utf-8')
        json_out = json.loads(info.decode(enc))
        return Response(json.dumps(json_out), status=200, headers={"content-type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/relays', methods=['GET'])
@apiKeyRequired
def relaysState():
    try:
        r = HTTPRequestsAPI.get('http://127.0.0.1:50500/api/relays/get')
        info = r.read()
        enc = r.info().get_content_charset('utf-8')
        json_out = json.loads(info.decode(enc))
        return Response(json.dumps(json_out), status=200, headers={"content-type": "application/json"})

    except Exception as e:
        log.error(e)
        return Response(status=500)


@bp.route('/api/v1/relay', methods=['PUT'])
@apiKeyRequired
def setRelay():
    try:
        cmd = json.loads(request.data)
        keys = list(cmd.keys())
        if len(keys) != 1:
            return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})

        if "relay" not in keys[0]:
            return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})

        name,relayId = keys[0].split("relay")
        if name != '' or not relayId.isnumeric():
            return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})

        if cmd[keys[0]] not in ["0", "1"]:
            return Response(json.dumps({"result":Result.INV_PARAM}),status=400, headers={"Content-Type": "application/json"})
        import urllib
        try:
            r = HTTPRequestsAPI.put('http://127.0.0.1:50500/api/relay/set',cmd)
            if r.status == 200:
                return Response(json.dumps({"result": Result.OK}), status=200, headers={"Content-Type": "application/json"})
        except urllib.error.HTTPError as e:
            answer = e.__dict__
            if answer['code'] == 403:
                return Response(json.dumps({"result": Result.NOT_ALLOWED}), status=403, headers={"Content-Type": "application/json"})
            if answer['code'] == 400:
                return Response(json.dumps({"result": Result.INV_PARAM}), status=400, headers={"Content-Type": "application/json"})
            if answer['code'] == 500:
                return Response(status=500)
        return Response(status=500)
    except Exception as e:
        log.error(e)
        return Response(f"{e}",status=400)
