import json
import logging
import time
import urllib.error
from flask import Blueprint, Response, request

from uci import UciException
from pathlib import Path

from barix.web.AES67 import stopBacoAudioLoop, startBacoAudioLoop, stopPlayingAES67, getSinkStatusAndErrors, getSourcesFromFile, storeSourcesInFile, sourcesMatch, playSource, AES67_DIR, checkIfCanPlaySource, SourceNotFoundException, SourceNotAliveException, SourceNotReadyException
from barix.web.uci import setUciConfigs
from barix.web.AES67_rest import AES67DaemonREST, BacoAudioLoopREST
from barix.web.auth import login_required


log = logging.getLogger('flask-backend')

bp = Blueprint('AES67_API', __name__)

p = Path(AES67_DIR)
if not p.exists():
    try:
        p.mkdir()
    except Exception as e:
        log.error(e)

@bp.route('/api/rest/v1/aes67/sources', methods=['GET'])
@login_required
def getAES67Sources():
    try:
        sourcesList = getSourcesFromFile()
    except Exception as e:
        log.error(e)
        return Response(status=500)
    else:
        return Response(json.dumps(sourcesList), status=200)

@bp.route('/api/rest/v1/aes67/config', methods=['PUT'])
@login_required
def saveAES67Configs():
    try:
        aes67Configs = json.loads(request.data)
    except Exception as e:
        log.error(e)
        return Response(status=400)

    if "ucis" in aes67Configs.keys():
        try:
            uciList = aes67Configs["ucis"]
            ucisConfigured = setUciConfigs(uciList, restartServices=False)
        except UciException as e:
            return Response("Invalid " + e, status=400)
        except Exception as e:
            log.error(e)
            return Response(status=500)
        else:
            try:
                # set new config in daemon if any uci changed
                if len(ucisConfigured) != 0:
                    mapping = {
                        'aes67.audio.playout_delay': 'playout_delay',
                        'aes67.audio.tic_frame_size_at_1fs': 'tic_frame_size_at_1fs',
                        'aes67.audio.sample_rate': 'sample_rate',
                        'aes67.network.rtsp_port': 'rtsp_port',
                        'aes67.network.sap_mcast_addr': 'sap_mcast_addr',
                        'aes67.network.sap_interval': 'sap_interval',
                        'aes67.network.mdns_enabled': 'mdns_enabled',
                        'aes67.network.ptp_domain': 'ptp_domain'
                    }
                    payload = {}
                    for uci in uciList:
                        payload[mapping[uci]] = uciList[uci]


                    log.info("Stop Baco audio loop to set new configs")
                    r = stopBacoAudioLoop()
                    if r != 201:
                        return Response(status=500)

                    aes67 = AES67DaemonREST()
                    result = aes67.setDaemonConfig(payload)
                    if result != 200:
                        return Response(status=500)
                    log.info("New configs set in aes67 daemon")

                    # start Baco Audio Loop
                    log.info("Start Baco Audio Loop again")
                    time.sleep(2)
                    r = startBacoAudioLoop()
                    if r != 201:
                        return Response(status=500)
            except Exception as e:
                log.error(e)
                return Response(status=500)

    if "sources" in aes67Configs.keys():
        # save AES67Sources in file
        try:
            storeSourcesInFile(aes67Configs["sources"])
        except Exception as e:
            log.error(e)
            return Response(status=500)

    return Response(status=200)

@bp.route('/api/rest/v1/aes67/status', methods=['GET'])
@login_required
def appIsReady():
    try:
        audio_loop = BacoAudioLoopREST(8088)
        audio_loop.getStats()
        #statsJson = audio_loop.getStats()
        #log.debug("Audio Loop Stats: {}".format(statsJson))
        #if statsJson["mode"] == "LOOP":
        #if statsJson["loopMode"] == 1:
        return Response(status=200)
    except urllib.error.URLError: # <urlopen error [Errno 111] Connection refused>
        return Response(status=503)  # Service Unavailable
    except Exception as e:
        log.error(e)
        return Response(status=500)



"""
sink parameters
{
  "name": "ALSA Sink 0",
  "io": "Audio Device",
  "delay": 576,
  "use_sdp": false,
  "source": "http://127.0.0.1:8080/api/source/sdp/0",
  "sdp": "v=0\no=- 0 0 IN IP4 127.0.0.1\ns=ALSA Source 0\nc=IN IP4 239.1.0.1/15\nt=0 0\na=clock-domain:PTPv2 0\nm=audio 5004 RTP/AVP 98\nc=IN IP4 239.1.0.1/15\na=rtpmap:98 L16/44100/2\na=sync-time:0\na=framecount:48\na=ptime:1.08843537415\na=mediaclk:direct=0\na=ts-refclk:ptp=traceable\na=recvonly\n",
  "ignore_refclk_gmid": false,
  "map": [0, 1]
}
"""
@bp.route('/api/rest/v1/aes67/sink', methods=['POST'])
@login_required
def playAES67Source():
    try:
        sourceReceived = json.loads(request.data)
    except Exception as e:
        log.error(e)
        return Response(status=400)
    try:
        sourceToPlay = checkIfCanPlaySource(sourceReceived)
        playSource(sourceToPlay)
    except SourceNotFoundException:
        log.error("SourceNotFoundException")
        return Response(status=404)
    except SourceNotAliveException:
        log.error("SourceNotAliveException")
        return Response("Not alive", status=400)
    except SourceNotReadyException:
        log.error("SourceNotReadyException")
        return Response("Not ready to play", status=400)
    except Exception as e:
        log.error(e)
        return Response(status=500)
    else:
        return Response(status=200)


@bp.route('/api/rest/v1/aes67/sink', methods=['DELETE'])
@login_required
def stopAES67Source():
    try:
        stopPlayingAES67()
    except Exception as e:
        log.error(e)
        return Response(status=500)
    else:
        return Response(status=200)


@bp.route('/api/rest/v1/aes67/ptp/config', methods=['PUT'])
@login_required
def setPTPConfig():
    try:
        ptpConfigs = json.loads(request.data)
    except Exception as e:
        log.error(e)
        return Response(status=400)
    try:
        ucisConfigured = setUciConfigs(ptpConfigs, restartServices=False)
    except UciException as e:
        return Response("Invalid " + e, status=400)
    except Exception as e:
        log.error(e)
        return Response(status=500)
    else:
        if len(ucisConfigured) != 0:
            try:
                payload = {"domain": ptpConfigs['aes67.network.ptp_domain'], "dscp": 48}
                aes67Daemon = AES67DaemonREST()
                result = aes67Daemon.setPTPConfig(payload)
                if result != 200:
                    return Response(status=500)
            except Exception as e:
                log.error(e)
                return Response(status=500)
    return Response(status=200)


@bp.route('/api/rest/v1/aes67/ptp/status', methods=['GET'])
@login_required
def getPTPStatus():
    aes67 = AES67DaemonREST()
    try:
        result = aes67.getPTPStatus()
        if type(result) is dict:
            return Response(json.dumps(result), status=200)
        else:
            return Response(status=500)
    except Exception as e:
        log.error(e)
    return Response(status=500)


@bp.route('/api/rest/v1/aes67/sink/status', methods=['GET'])
@login_required
def handleGetRTPSinkStatus():
    try:
        status = getSinkStatusAndErrors(0) #we're going to have one only
        #log.debug("Sink status and errors: {}".format(status))
        if len(status.keys()) > 0:
            aes67 = AES67DaemonREST()
            result = aes67.getAllRTPSinks()
            if type(result) is dict:
                if len(result['sinks']) > 0:
                    for sink in result['sinks']:
                        if sink['id'] == 0:
                            status['source_id'] = sink['name'].replace('Sink ', '')
                            fileSources = getSourcesFromFile()
                            #log.debug(fileSources)
                            for source in fileSources:
                                if source['id'] == status['source_id']:
                                    if source['alias'] == '':
                                        status['source_name'] = source['name']
                                    else:
                                        status['source_name'] = source['alias']
        return Response(json.dumps(status), status=200)
    except Exception as e:
        log.error(e)
        return Response(status=500)
