#!/usr/bin/python3
import logging
import Logger
import time
import subprocess
import signal
import sys
import os
from threading import *
from flask import Flask
from hwType import *
import flexa_uci
from GenericRestAPI import *
from barix.io import leds

class AES67_Mode:
    HTTP_API_IPADDRES   = "127.0.0.1"
    HTTP_API_PORT       = 50500

    def __init__(self, hwType):
        self.hwType         = hwType
        self.sampleRate     = 44100
        self.bufferSize     = 192
        self.periodSize     = 32
        self.pcmDevice      = None
        self.bacoProc       = None
        self.aes67API       = None
        self.stopRunning    = False
        self.isMuted        = True
        self.appPath        = os.path.dirname(os.path.realpath(__file__))
        self.led            = leds.IpamLed("led2")
        self.bacoAPI        = GenericRestAPI(8088)
        self.aes67DaemonAPI = GenericRestAPI(8080)
        self.ampAPI         = GenericRestAPI(50555)
        Logger.initLogger('/var/log/ipac.log', 2 * 1024 * 1024, 3, False)
        self.logger = logging.getLogger(__name__)
        self.logger.info("IP Audio Client Starting -- aes67 mode")
        signal.signal(signal.SIGINT, self.signalHandler)
        signal.signal(signal.SIGTERM, self.signalHandler)

    def signalHandler(self, sig, frame):
        self.logger.info('SIGINT or SIGTERM received')
        self.stopRunning = True

    """
    Start services
    """
    def start(self,app):
        self.logger.info('Start services in AES67 Mode...')
        self.selectPcmDevice()

        """
        remove status.json file used by aes67-daemon
        """
        if os.path.exists("/mnt/data/aes67/status.json"):
            os.remove("/mnt/data/aes67/status.json")

        self.runService("/etc/init.d/aes67-demo start", wait=True)
        self.logger.info('Starting AES67 daemon')

        # disable headphones/RCA output
        self.runService("echo 14 >/sys/class/gpio/export", wait=True)
        self.runService("echo out > /sys/class/gpio/gpio14/direction", wait=True)

        self.logger.info('kill aplay')
        self.runService("killall -9 qiba-aplay-loop.sh", wait=True)
        self.runService("killall -9 aplay", wait=True)

        self.startBacoAudioLoop(log="yes", csv="no")

        # enable headphones/RCA output
        self.runService("echo in > /sys/class/gpio/gpio14/direction", wait=True)

        self.led.set("green")

        if self.hwType == HwType.BARIX_TPA400:
            time.sleep(0.5)
            res = self.ampAPI.post('/api/amp/init', '')
            self.logger.info("Initialize TPA400 tas2770 amplifier, api retcode {}".format(res))

        self.aes67API = app
        self.runFlaskThread()

        self.runMainLoop()

    """ 
    Select pcm device based on board HW Type
    Digital Out boards: TPA400, ECM400 and MPA400 (for the moment MPA only uses digital)
    Analog out: default
    """
    def selectPcmDevice(self):
        if self.hwType == HwType.BARIX_TPA400 or self.hwType == HwType.USSI_ENC_CONTROLLER or self.hwType == HwType.BARIX_MPA400 or self.hwType == HwType.HYPERSPIKE_MPA400:
            self.pcmDevice = "hw:1,0"
        else:  # analog_out
            self.pcmDevice = "hw:2,0"

    def startBacoAudioLoop(self, log=None, csv=None):
        cmd = f"baco-audio-loop --output={self.pcmDevice} --rate={self.sampleRate} --buffer-size={self.bufferSize} --period-size={self.periodSize} --vol={0}"
        if log == "yes":
            cmd = cmd + " --log-file=/var/log/audio-loop.log"
        if csv == "yes":
            cmd = cmd + " --csv-file=/var/log/delay.csv"
        self.logger.info("Starting baco-audio-loop in silence mode")
        self.bacoProc = self.runService(cmd, wait=False)
        time.sleep(3)
        baco_running = {"mode": "LOOP",
                        "parameters": {"device": "hw:3,0", "sampleRate": 48000, "fifoDelay": 92, "dispatchSize": 16}}
        res = self.bacoAPI.put('/api/v1/input/config', baco_running)
        self.logger.info("Set baco-audio-loop into Running Mode - status code: {}".format(res))
        self.isMuted = True

    def stopBacoAudioLoop(self):
        if self.bacoProc is not None:
            self.bacoProc.terminate()
        self.bacoProc = None

    def runService(self, cmd, wait=False):
        p = subprocess.Popen(cmd, shell=True)
        if wait:
            p.wait()
        return p

    def runFlaskThread(self):
        flaskBaco_thread = Thread(target=self.runFlask)
        flaskBaco_thread.daemon = True
        flaskBaco_thread.start()

    def runFlask(self):
        self.logger.info("Running Internal API")
        self.aes67API.run(host=AES67_Mode.HTTP_API_IPADDRES, port=AES67_Mode.HTTP_API_PORT)

    def setAES67Daemon(self):
        try:
            ptpDomain = int(flexa_uci.uci_get("aes67.network.ptp_domain"))
            sapMcastAddr = flexa_uci.uci_get("aes67.network.sap_mcast_addr")
            if flexa_uci.uci_get("aes67.network.mdns_enabled") == 'true':
                mdnsEnabled = True
            else:
                mdnsEnabled = False
            payload = {"rtsp_port": 8854, "playout_delay": 0, "tic_frame_size_at_1fs": 48, "sample_rate": 48000,
                       "rtp_mcast_base": "239.1.0.1", "rtp_port": 5004, "ptp_domain": ptpDomain,
                       "ptp_dscp": 48, "sap_mcast_addr": sapMcastAddr, "sap_interval": 30, "mdns_enabled": mdnsEnabled}
            self.aes67DaemonAPI.post('/api/config', payload)
        except Exception as e:
            self.logger.error(f"Set Aes67 Daemon config: {e}")

    def checkForUnderrunsState(self):
        try:
            res = self.bacoAPI.get('/api/v1/stats')
            if "loopMode" in res:
                if res["loopMode"] == 2 and res['stats']["xrunsOut"] >= 3:
                    self.logger.warning("UNDERRUN state => Restart AES67 Daemon and kill baco-audio-loop")
                    self.stopBacoAudioLoop()
                    self.setAES67Daemon()
                    time.sleep(2)
                    self.startBacoAudioLoop("yes")
                    return True
        except Exception as e:
            self.logger.error(f"Exception: {e}")
        finally:
            return False

    def checkAudioLoopTriggerCondition(self):
        try:
            res = self.bacoAPI.get('/api/v1/stats')
            if res["mode"] == "LOOP":
                if res["loopMode"] == 1 and res["stats"]["level"] <= 300:
                    return True
            return False
        except Exception as e:
            self.logger.error(f"Getting AudioLoop stats: {e}")
            return False

    def checkAmp(self):
        """
        check if TPA tas2770 is in Software Shutdown mode
        This check should be done when audioloop is in Loop Running mode
        """
        try:
            res = self.ampAPI.get('/api/amp/mode')
            if res["mode"] == 2:
                self.logger.info("TAS2770 amp is in Software Shutdown mode")
                res = self.ampAPI.post('/api/amp/init', '')
                self.logger.info("Initialize TPA400 tas2770 amplifier, api retcode {}".format(res))
        except :
            self.logger.error("check tas2770 amp failed")

    def setMainVol(self, vol):
        self.bacoAPI.put('/api/v1/volume', {"volume": vol})
        self.logger.info(f"Set Main Volume to {vol}%")

    def runMainLoop(self):
        while not self.stopRunning:
            if not self.checkForUnderrunsState():
                if self.checkAudioLoopTriggerCondition():
                    if self.hwType == HwType.BARIX_TPA400:
                        self.checkAmp()
                    muted = flexa_uci.uci_get('application.audio.muted')
                    if self.isMuted and muted == "false":
                        vol = flexa_uci.uci_get('application.audio.volume')
                        self.setMainVol(vol)
                        self.isMuted = False
            time.sleep(1)
        self.led.set("red")
        self.logger.info('shutting down Flask WebUI')
        self.logger.info('shutting down baco-audio-loop')
        self.stopBacoAudioLoop()
        self.logger.info('shutting down aes67-daemon')
        self.runService("/etc/init.d/aes67-demo stop", wait=True)
        sys.exit(0)


if __name__ == "__main__":

    if len(sys.argv) != 2:
        print("need to receive hwType argument")
        sys.exit(1)

    hwType = int(sys.argv[1])
    print("Running on device type:{}".format(hwType))

    aes67APP = AES67_Mode(hwType)

    app = Flask(__name__)

    @app.route('/api/baco/start', methods=['POST'])
    def baco_start():
        aes67APP.startBacoAudioLoop(log="yes", csv="no")
        return "Start Baco Audio Loop", 201

    @app.route('/api/baco/stop', methods=['POST'])
    def baco_stop():
        aes67APP.stopBacoAudioLoop()
        return "Stop Baco Audio Loop", 201

    aes67APP.start(app)
