//
// Created by to on 24/02/21.
//


#include <baco/audio/SilenceSource.h>
#include <baco/ClockMonotonic.h>

#include "AudioLoopCore.h"

using namespace baco;
using namespace Poco;
using namespace std;

AudioLoopCore::AudioLoopCore( baco::AlsaSinkPtr const& sink , baco::FrameFormat const& outFmt, int compensationInterval, double volume, baco::FrameReplicatorPtr const& replicator):
logger{ Logger::get("loop-core") } ,
myOutFmt{ outFmt } ,
mySilenceSource{ make_unique<baco::SilenceSource>(outFmt,32) } ,
myAlsaSink{ sink } ,
myCompensationInterval{ compensationInterval } ,
myVolume{ volume },
myFrameReplicator {replicator},
myExecutor{1,1}
{
    if ( myCompensationInterval>0 ) {
        myExecutor.schedulePeriodic([this] () {this->periodComp();},myCompensationInterval,myCompensationInterval);
    }
}


void AudioLoopCore::setCsvReport( std::string const& fileName )
{
    if ( fileName.empty() ) {
        return;
    }

    myCsvOfs.open(fileName);
    if ( myCsvOfs.is_open() ) {
        log_information(logger,"opened CSV report.");
        myCsvOfs << "TIME(ms),FIFO_LEVEL(frames),LEVEL_MIN,LEVEL_MAX,XRUNS_IN,XRUNS_OUT,RESAMPLER_DELAY,SAMPLES_IN,SAMPLES_OUT,COMP,FIRST_XRUN,T_LV_MIN,T_LV_MAX\n";
        myCsvOfs.flush();
        myCsvEnabled  = true;
    }
    else {
        log_error(logger,"CSV report not open!");
    }
}


baco::ReturnCode AudioLoopCore::setInputLoop( InputBlockConfig const& loopCfg )
{
    std::unique_lock<mutex> lock{myMutex};
    myInputConfig       = loopCfg;
    myRequestedMode     = LoopMode::LOOP_STARTING;
    myCompAcc           = 0;
    myCompCount         = 0;
    myRequestPending    = true;
    return ReturnCode::OK;
}


baco::ReturnCode AudioLoopCore::disconnectInputLoop()
{
    std::unique_lock lock{myMutex};
    myRequestedMode     = LoopMode::SILENT;
    myRequestPending    = true;

    return ReturnCode::OK;
}


AudioLoopStats AudioLoopCore::getStats()
{
    AudioLoopStats loopStats{};
    std::unique_lock<mutex> lock{myMutex};
    loopStats.mode      = myCurrentMode;
    if ( myCurrentMode != LoopMode::SILENT){
        loopStats.input                 = InputBlockStats{myInputBlock->getStats()};
        loopStats.output.sampleCount    = myAlsaSink->getSampleCount() - myOutSampleCountOffset;
        loopStats.output.numXruns       = myOutXrun;
        loopStats.output.tFirstXrun     = myTimeFirstOutXrun;
    }
    return loopStats;
}

void AudioLoopCore::setVolume(int vol)
{
    std::unique_lock<mutex> lock{myMutex};
    myVolume = vol*0.01;
    log_information(logger,format("setVolume (%d): %.2f",vol, myVolume));
    if ( myCurrentMode != LoopMode::SILENT )
        myInputBlock->setGain(myVolume,0,0);
}

void AudioLoopCore::iterate()
{
    if ( myRequestPending ) {
        scoped_lock lock{myMutex};
        if ( myRequestedMode==LoopMode::SILENT ) {
            myInputBlock.reset();
            myCurrentMode = LoopMode::SILENT;
        }
        else if ( myRequestedMode==LoopMode::LOOP_STARTING ) {
            myInputBlock  = make_unique<InputBlock>(myInputConfig,myOutFmt);
            if ( myFrameReplicator == nullptr ){
                myInputBlock->attachSink(myAlsaSink);
            }
            else {
                myInputBlock->attachSink(myFrameReplicator);
            }
            myInputBlock->setGain(0,0,0);
            myCurrentMode = LoopMode::LOOP_STARTING;
            myOutSampleCountOffset  = myAlsaSink->getSampleCount();
            myTimeFirstOutXrun      = 0;
            myOutXrun               = 0;
            myCsvStartTime          = ClockMonotonic::getMillis();
            myAvgComp               = 0;
        }
        else {
            log_error(logger,format("unknown requested mode: %d",(int)myRequestedMode));
        }
        myRequestPending          = false;
        log_information(logger,"request fulfilled");
    }

    if ( myCurrentMode!=LoopMode::SILENT ) {
        if ( myClearStats ) {
            myInputBlock->clearStats();
            myOutXrun               = 0;
            myTimeFirstOutXrun      = 0;
            myOutSampleCountOffset  = myAlsaSink->getSampleCount();
            myClearStats            = false;
        }
        auto rc = myInputBlock->readFrame();
        if ( rc == ReturnCode::E_BADSTATE ) {
            myInputBlock->setGain(0,0,0);
            myCompCount = 0;
            myCurrentMode   = LoopMode::LOOP_STARTING;
            return;
        }
    
        int x       = myAlsaSink->getNumXruns();
        int diff    = x - myLastXrunCnt;
        if ( diff!=0 ) {
            if ( myOutXrun==0 ) {
                myTimeFirstOutXrun  = ClockMonotonic::getMillis() - myCsvStartTime;
            }
            myRecoveryAvg   = myAvgComp;
            log_information(logger, "Underrun ");
            if(myCurrentMode == LoopMode::LOOP_RUNNING) {
                myCurrentMode   = LoopMode::LOOP_RECOVERING;
                myCompAcc       = 0;
                myCompCount     = 0;
            }
            myOutXrun      += diff;
        }
        myLastXrunCnt       = x;
    }
    else {
        mySilenceSource->readFrame(myAlsaSink);
    }
}


void AudioLoopCore::periodComp()
{
    unique_lock lock{myMutex};
    int curComp = 0;
    log_debug(logger,format("Period Comp: Current Mode %d",(int)myCurrentMode)); 
    if ( myCurrentMode==LoopMode::LOOP_STARTING ) {
        curComp = myInputBlock->calculateCompensation();
        myInputBlock->applyCompensation(curComp);
        myCompAcc  += curComp;
        myCompCount++;
        if ( myCompCount==2 ) {
            double compAvg  = myCompAcc/2;
            if ( abs(compAvg) < 45 ) {
                log_information(logger,format("setVolume : %.2f", myVolume));
                myInputBlock->setGain(myVolume,0,0);
                myCurrentMode   = LoopMode::LOOP_RUNNING;
            }
            myCompAcc   = 0;
            myCompCount = 0;
            myAvgComp   = 0;
        }
    }
    else if ( myCurrentMode==LoopMode::LOOP_RUNNING ) {
        curComp = myInputBlock->calculateCompensation();
        myInputBlock->applyCompensation(curComp);
        myCompAcc  += curComp;
        myCompCount++;
        if ( myCompCount>=60 ) { // the > is just in case :)
            double avgComp  = myCompAcc/60;
            log_information(logger,format("compensation in period: AVERAGE(%.1f) %+.1f",myAvgComp,avgComp));
            myAvgComp      += avgComp;
            myInputBlock->setAvgComp(myAvgComp);
            myCompCount     = 0;
            myCompAcc       = 0;
        }
    }
    else if ( myCurrentMode==LoopMode::LOOP_RECOVERING ) {
        curComp = myInputBlock->calculateCompensation();
        myInputBlock->applyCompensation(curComp);
        myCompAcc  += curComp;
        myCompCount++;
        if ( myCompCount>=5 ) {
            double curDelta     = myCompAcc/5;
            double avgComp      = curDelta + myAvgComp;
            myCompAcc           = 0;
            myCompCount         = 0;
            log_information(logger, format("curDelta= %.1f, avgComp= %.1f, myRecoveryAvg = %.1f, abs(dif)= %.1f",curDelta, avgComp, myRecoveryAvg,  (double) abs(avgComp-myRecoveryAvg)));
            if ( abs(avgComp-myRecoveryAvg)<10 || abs(curDelta)<2 ) {
                log_information(logger,format("recovery complete. average compensation is %.1f samples/sec",avgComp));
                myCurrentMode   = LoopMode::LOOP_RUNNING;
                myInputBlock->setAvgComp(avgComp);
                myAvgComp       = avgComp;
            }
            myRecoveryAvg   = avgComp;
        }
    }

    if ( myCsvEnabled && myCurrentMode!=LoopMode::SILENT ) {
        int64_t t   = baco::ClockMonotonic::getMillis() - myCsvStartTime;
        assert( myInputBlock!=nullptr );
        InputBlockStats s{myInputBlock->getStats()};
        lock.unlock();

        myCsvOfs << format("%?d,%d,%d,%d,%d,%?u,%d,%?u,%?u",
                                t,
                                s.fifoLevel,
                                s.fifoLevelMin,
                                s.fifoLevelMax,
                                s.xrunIn,
                                myOutXrun ,
                                s.resamplerDelay,
                                s.totalSamplesIn ,
                                myAlsaSink->getSampleCount() - myOutSampleCountOffset )
                        << format(",%?d,%?u,%?u,%?u\n",
                                  curComp ,
                                  myTimeFirstOutXrun ,
                                  s.tLevelMin  - myCsvStartTime ,
                                  s.tLevelMax  - myCsvStartTime
                                  );
        myCsvOfs.flush();
    }
    myClearStats        = true;
}

